from typing import List, Union, Tuple
from akasha.utils.prompts.format import language_dict
from urllib.parse import urlparse
import os
# from langchain_core.messages import HumanMessage, SystemMessage, AIMessage

sys_s = "[INST] <<SYS>> "
sys_e = " <<SYS>> [/INST]\n\n"


def _separate_name(name: str):
    """separate type:name by ':'

    Args:
        **name (str)**: string with format "type:name" \n

    Returns:
        (str, str): res_type , res_name
    """
    sep = name.split(":")

    if len(sep) > 2:
        res_type = sep[0].lower()
        res_name = ":".join(sep[1:])
    elif len(sep) < 2:
        ### if the format type not equal to type:name ###
        res_type = sep[0].lower()
        res_name = ""
    else:
        res_type = sep[0].lower()
        res_name = sep[1]

    return res_type, res_name


def format_chat_gemini_prompt(system_prompt: str, prompt: str) -> List[dict]:
    if system_prompt == "" and prompt == "":
        return []

    if system_prompt == "":
        return [{"role": "user", "parts": [prompt]}]

    if prompt == "":
        return [{"role": "model", "parts": [system_prompt]}]

    return [
        {"role": "model", "parts": [system_prompt]},
        {"role": "user", "parts": [prompt]},
    ]


def format_chat_gemma_prompt(system_prompt: str, prompt: str) -> List[dict]:
    if system_prompt == "" and prompt == "":
        return []

    if system_prompt == "":
        return [{"role": "user", "content": prompt}]

    if prompt == "":
        return [{"role": "user", "content": system_prompt}]

    return [{"role": "user", "content": system_prompt + "\n\n" + prompt}]


def format_chat_gpt_prompt(system_prompt: str, prompt: str) -> List[dict]:
    if system_prompt == "" and prompt == "":
        return []

    if system_prompt == "":
        return [{"role": "user", "content": prompt}]

    if prompt == "":
        return [{"role": "system", "content": system_prompt}]

    return [
        {"role": "system", "content": system_prompt},
        {"role": "user", "content": prompt},
    ]


def format_chat_mistral_prompt(system_prompt: str, prompt: str) -> List[dict]:
    if system_prompt == "" and prompt == "":
        return []

    if system_prompt == "":
        return [{"role": "user", "content": prompt}]

    if prompt == "":
        return [
            {"role": "user", "content": "start conversation."},
            {"role": "assistant", "content": system_prompt},
        ]

    return [
        {"role": "user", "content": "start conversation."},
        {"role": "assistant", "content": system_prompt},
        {"role": "user", "content": prompt},
    ]


def format_llama_sys_prompt(system_prompt: str, prompt: str) -> str:
    if system_prompt == "" and prompt == "":
        return ""

    if system_prompt == "":
        return "[INST] " + prompt + " [/INST]\n"
    if prompt == "":
        return "[INST] <<SYS>> " + system_prompt + " <<SYS>> \n"

    return "[INST] <<SYS>> " + system_prompt + " <<SYS>> \n\n" + prompt + " [/INST]\n"


def format_GPT_sys_prompt(system_prompt: str, prompt: str) -> str:
    if system_prompt == "" and prompt == "":
        return ""

    if system_prompt == "":
        return "Human: " + prompt + "\n\n"
    if prompt == "":
        return "System: " + system_prompt + "\n"

    return "System: " + system_prompt + "\n\n" + "Human: " + prompt + "\n"


def format_sys_prompt(
    system_prompt: str,
    prompt: str,
    prompt_format_type: str = "auto",
    model: str = "remote:xxx",
) -> Union[str, List[dict]]:
    if prompt_format_type.lower() == "auto":
        prompt_format_type = decide_auto_prompt_format_type(model)

    if prompt_format_type.lower() == "llama":
        ret_text = format_llama_sys_prompt(system_prompt, prompt)

    elif prompt_format_type.lower() == "chat_gpt":
        ret_text = format_chat_gpt_prompt(system_prompt, prompt)

    elif prompt_format_type.lower() == "chat_mistral":
        ret_text = format_chat_mistral_prompt(system_prompt, prompt)

    elif prompt_format_type.lower() == "chat_gemma":
        ret_text = format_chat_gemma_prompt(system_prompt, prompt)

    elif prompt_format_type.lower() == "chat_gemini":
        ret_text = format_chat_gemini_prompt(system_prompt, prompt)

    else:
        ret_text = format_GPT_sys_prompt(system_prompt, prompt)

    return ret_text


def decide_auto_prompt_format_type(model: str = "remote:xxx") -> str:
    """since prompt_format_type is "auto", so we based on the model name to decide which prompt format type to use

    Args:
        model (str, optional): model_type and model name. Defaults to "remote:xxx".

    Returns:
        str: _description_
    """
    model_type, model_name = _separate_name(model)

    if "openai" in model_type or "openai" in model_name:
        return "chat_gpt"
    elif "gemini" in model_type or model_type in ["google", "gemi"]:
        return "chat_gemini"
    elif (
        "anthropic" in model_type
        or "claude" in model_name
        or "mistral" in model_name
        or "mixtral" in model_name
        or "gemma" in model_name
    ):
        return "chat_mistral"

    elif "llama" in model_type or "llama" in model_name:
        return "llama"

    return "gpt"


def format_history_prompt(
    history_messages: list,
    prompt_format_type: str = "chat_gpt",
    user_tag="user",
    assistant_tag="assistant",
) -> List[dict]:
    prod_history = []
    if isinstance(history_messages, str):
        history_messages = [history_messages]

    for idx, msg in enumerate(history_messages):
        if idx % 2 == 0:
            prod_history.append({"role": user_tag, "content": msg})
        else:
            prod_history.append({"role": assistant_tag, "content": msg})

    return prod_history


def format_question_query(question: list, answer: str) -> Tuple[str, str]:
    """generate a certain format of question to input to llm. Last element means which selection is the correct answer.
       return the question query string and the answer string.\n
      example:    ["what is 1+1 euqals to?", "2", "4", "8", "10", "1"]
      after :
        query = " what is 1+1 euqals to?
            1. 2
            2. 4
            3. 8
            4. 10
            "
        ans = "1"
    Args:
        **question (list)**: list of question and selections\n

    Returns:
        (str, str): return the question query string and the answer string
    """
    n = len(question)
    if n == 0:
        return "", ""
    elif n == 1:
        return "Question: " + question[0], ""
    query = "Question: " + question[0] + "\n"

    for i in range(1, n):
        query += question[i] + "\n"

    return query, answer


def format_llama_json(query: str):
    """insert system prompt for llm to generate JSON format of {"ans":selection number}

    Args:
        **query (str)**: a question string with selections that we want llm to answer\n

    Returns:
       prompt (str): the whole prompt includes system prompt and question
    """

    sys_prompt = (
        "human will give you a question with several possible answer, use the content of documents "
        + "to choose correct answer. and you need to return the answer of this question as"
        + ' JSON structure with key "ans", and only this JSON structure, please don\'t add any other word. for example, '
        + 'User: what is 1+1 euqals to? 1.「2」  2.「4」 3.「10」 4.「15」 \n you: {"ans":1}'
    )

    prompt = sys_s + sys_prompt + sys_e
    return prompt + query


def format_chinese_json(query: str):
    """system prompt for single choice question in chinese

    Args:
        **query (str)**: question string\n

    Returns:
        str: combined prompt
    """
    sys_prompt = (
        "### 指令: 我会给出一个问题和几个可能的选项，请只根据提供的文件找到其中正确的一个答案，"
        + "並回答答案為第幾個選項。若沒有提供，請照你的知識回答，并将答案以JSON的格式表示，如答案為第一個選項，"
        + "回答的格式為{'ans':1}，不要添加其他字。  ### 问题和选项:\n"
    )
    # ts_prompt = "### Instruction: I will provide a question and several possible options in the input. Please find the correct answer based solely on the provided texts, and respond with the number of the option that is the correct answer. If no texts is provided, please respond based on your knowledge, and format the answer in JSON format. For example, if the answer is the first option, the format of the response should be {'Answer': 1}. Please do not add any additional words. ### Input:"
    # sys_prompt = ts_prompt + query "  ### Response:"
    return sys_prompt + query


def format_wrong_answer(
    num: int, doc_text: str, question: str, correct_ans: str
) -> str:
    """prompt for generate wrong answers to create single choice question

    Args:
        **num (int)**: number of wrong answers that we want to generate\n
        **doc_text (str)**: document texts that used to generate question\n
        **question (str)**: question string\n
        **correct_ans (str)**: correct answer string\n

    Returns:
        str: combined prompt
    """

    q_prompt = (
        sys_s
        + f"根據以下的文件、問題和正確答案，請基於文件、問題和正確答案生成{num}個錯誤答案，錯誤答案應該與正確答案有相關性但數字、內容或定義錯誤，或者與正確答案不相同但有合理性。並注意各個錯誤答案必須都不相同。\n\n示例格式：\n<開始文件>\n...\n<結束文件>\n<開始問題>\n...\n<結束問題>\n<開始正確答案>\n...\n<結束正確答案>\n\n錯誤答案：錯誤答案1在這里\n\n錯誤答案：錯誤答案2在這里\n\n錯誤答案：錯誤答案3在這里\n\n。開始吧！"
        + sys_e
        + "<開始文件>\n"
    )
    end_doc = "<結束文件>\n"
    st_q = "<開始問題>\n"
    end_q = "<結束問題>\n"
    st_cor = "<開始正確答案>\n"
    end_cor = "<結束正確答案>\n"
    q_prompt = (
        q_prompt
        + doc_text
        + end_doc
        + st_q
        + question
        + end_q
        + st_cor
        + correct_ans
        + end_cor
        + "\n\n"
    )

    return q_prompt


def format_create_question_prompt(
    doc_text: str,
    question_type: str = "fact",
    question_style: str = "essay",
    topic: str = "",
) -> str:
    """prompts for auto generate question from document

    Args:
        **doc_text (str)**: texts from documents\n
        **question_type (str)**: question type, can be "single choice", "essay"\n

    Returns:
        str: combined prompt
    """
    qt = ""
    question_type = question_type.lower()
    question_style = question_style.lower()
    if question_style not in ["essay", "問答", "essay question", "essay_question"]:
        qt = "少於100字的"

    if question_type in ["fact", "facts", "factoid", "factoids", "事實"]:
        q_prompt = fact_question_prompt(qt, topic)
    elif question_type in [
        "summary",
        "sum",
        "summarization",
        "summarize",
        "summaries",
        "摘要",
    ]:
        q_prompt = summary_question_prompt(qt)
    elif question_type in ["irre", "irrelevant", "irrelevance", "無關"]:
        q_prompt = irre_question_prompt(qt, topic)

    # end_prompt = "<End Document>\n"
    end_prompt = "<結束文件>\n"
    # generate question prompt = generate_question_prompt(Document)
    q_prompt = q_prompt + doc_text + end_prompt

    return q_prompt


def fact_question_prompt(qt: str = "", topic: str = "") -> str:
    """prompt for generate fact question"""

    if topic != "":
        topic = f"相關{topic}的"

    fact_prompt = f"人類：您是一位教師，正在為測驗準備問題。\n請基於文件只生成一個{topic}問題和一個{qt}答案，問題應該詳細並且明確基於文件中的訊息，但不要使用'根據文件中的訊息'等詞語取代想詢問的問題內容。"
    format_prompt = "\n\n示例格式：\n<開始文件>\n...\n<結束文件>\n問題：問題在這里\n答案：答案在這里\n\n。開始吧！"
    return sys_s + fact_prompt + format_prompt + sys_e + "<開始文件>\n"


def summary_question_prompt(qt: str = "") -> str:
    """prompt for generate summary question"""

    sum_prompt = '人類： 請根據以下文件進行摘要，並將摘要放在 "答案"後面.'
    format_prompt = (
        "\n\n示例格式：\n<開始文件>\n...\n<結束文件>\n答案：摘要在這裡\n\n。開始吧！"
    )
    return sys_s + sum_prompt + format_prompt + sys_e + "<開始文件>\n"


def irre_question_prompt(qt: str = "", topic: str = "") -> str:
    """prompt for generate irrelevant question"""
    if topic != "":
        topic = f"和{topic}"

    irre_prompt = f"人類：我想測試語言模型根據文件回答的可靠性。\n請只生成一個名詞或領域與文件內容{topic}相關，但文件內容中不存在的問題。這代表說，詢問任何人這個問題與文件，都無法回答該問題。問題應該詳細但不要使用'根據文件中的訊息'等詞語取代想詢問的問題內容。根據這個問題和文件，生成一個{qt}回答"
    format_prompt = "\n\n示例格式：\n<開始文件>\n...\n<結束文件>\n問題：問題在這里\n答案：答案在這里\n\n。開始吧！"
    return sys_s + irre_prompt + format_prompt + sys_e + "<開始文件>\n"


def compare_question_prompt(
    question_style: str, topic: str, nouns: str, used_texts: str
):
    """prompt for generate compare question"""

    qt = ""
    if question_style not in ["essay", "問答", "essay question", "essay_question"]:
        qt = "少於100字的"

    com_prompt = f"人類：您是一位教師，正在為測驗準備問題。\n請基於文件和相關主題只生成一個相關主題之間的比較問題和一個{qt}答案，問題應該詳細並且明確基於文件中的訊息，但不要使用'根據文件中的訊息'等詞語取代想詢問的問題內容。"
    format_prompt = "\n\n示例格式：\n<開始文件>\n...\n<結束文件>\n<開始相關主題>\n...\n<結束相關主題>\n問題：問題在這里\n答案：答案在這里\n\n。開始吧！"

    return (
        sys_s
        + com_prompt
        + format_prompt
        + sys_e
        + "<開始文件>\n"
        + used_texts
        + "<結束文件>\n"
        + "<開始相關主題>\n"
        + nouns
        + "<結束相關主題>\n"
    )


def format_llm_score(cand: str, ref: str):
    """the system prompt for llm to calculate the cnadidate is correct or not.

    Args:
        **cand (str)**: llm generated response that we want to test the performance\n
        **ref (str)**: reference answer\n
    """

    sys_prompt = (
        "human will give you a [candidate] sentence and a [reference] sentence, please score the [candidate] sentence "
        + "based on the [reference] sentence, the higher score means the [candidate] sentence has enough information and correct answer that [reference] sentence has."
        + "remember, you can only return the score and need to return the score of this [candidate] sentence as a float number range from 0 to 1.\n"
        + "Example Format:\n Human: [candidate]: ...\n\n [reference]: ...\n\n You: 0.8\n\n"
    )

    # prompt = sys_s + sys_prompt + sys_e
    return sys_prompt, "[candidate]: " + cand + "\n\n[reference]: " + ref + "\n\n"


def format_reduce_summary_prompt(
    cur_text: str, summary_len: int = 500, language: str = "zh"
):
    """the prompt for llm to generate a summary of the given text

    Args:
        **cur_text (str)**: the text that we want llm to generate a summary\n
        **summary_len (int, optional)**: the summary word length we want llm to generate. Defaults to 500.\n

    Returns:
        str: summary prompt.
    """
    target_language = language_dict[language]
    underline = "------------"

    if summary_len > 0:
        if "chinese" in target_language:
            sys_prompt = (
                f"將以下內容總結成一個{summary_len}字的摘要，一步一步來，如果你做得好，我會給你100美元小費。\n\n"
                + underline
                + "\n"
                + cur_text
                + underline
            )
        else:
            sys_prompt = (
                f"Write a concise {summary_len} words summary of the following text. Do it step by step and I will tip you 100 bucks if you are doing well.\n\n"
                + underline
                + "\n"
                + cur_text
                + underline
            )

    else:
        if "chinese" in target_language:
            sys_prompt = (
                "將以下內容總結成一個摘要，一步一步來，如果你做得好，我會給你100美元小費。\n\n"
                + underline
                + "\n"
                + cur_text
                + underline
            )
        else:
            sys_prompt = (
                "Write a concise summary of the following. Do it step by step and I will tip you 100 bucks if you are doing well.\n\n"
                + underline
                + "\n"
                + cur_text
                + underline
            )

    return sys_prompt


def format_refine_summary_prompt(
    cur_text: str, previous_summary: str, summary_len: int = 500, language: str = "zh"
):
    """the prompt for llm to generate the summary of the given text and previous summary

     Args:
        **cur_text (str)**: the text that we want llm to generate a summary\n
        **previous_summary (str)**: the previous summary that we want llm to generate a summary\n
        **summary_len (int, optional)**: the summary word length we want llm to generate. Defaults to 500.\n

    Returns:
        str: summary prompt.

    """
    target_language = language_dict[language]
    if "chinese" in target_language:
        sys_prompt = f"""將以下內容總結成一個{summary_len}字的摘要。
        我們已經提供了一個現有的摘要，原始摘要如下：{previous_summary}
        ------------\n
        {cur_text}\n
        ------------\n
        根據新的內容，修改原始摘要。
        如果新的內容不夠有用，請返回原始摘要。
        一步一步來，如果你做得好，我會給你100美元小費。
        """
    else:
        sys_prompt = f"""Your job is to produce a final summary of {summary_len} words.
        We have provided an existing summary up to a certain point, original summary is:  {previous_summary}
        ------------\n
        {cur_text}\n
        ------------\n
        Given the new context, refine the original summary.
        If the context isn't useful, return the original summary.
        Do it step by step and I will tip you 100 bucks if you are doing well.
        """
    return sys_prompt


def format_compression_prompt(query: str, doc: str):
    return f"""Given the following question and context, extract any part of the context *AS IS* that is relevant to answer the question. If none of the context is relevant return an empty string. 

    Remember, *DO NOT* edit the extracted parts of the context.
    \nQuestion: {query}
    \nContext:
    {doc}"""


def format_pic_summary_prompt(chunk_size: int = 500):
    # f"please use traditional chinese to describe this picture in {chunk_size} words.\n\n"
    return "please use traditional chinese to describe this picture in details.\n\n"


def default_doc_ask_prompt(language: str = "zh"):
    prompt = """Use the following pieces of context to answer the user's question. 
If you don't know the answer, just say that you don't know, don't try to make up an answer.\n"""
    if "chinese" in language_dict[language]:
        prompt += "用中文回答下列問題：\n"

    return prompt


def default_ask_prompt(language: str = "zh"):
    prompt = """你是擅長回答問題的助手，若有提供參考文件，只根據參考文件回答問題;若沒有參考文件，則根據你的理解回答使用者的問題。\n"""

    if "chinese" in language_dict[language]:
        prompt += "用中文回答下列問題：\n"

    return prompt


def default_conclusion_prompt(question: str, language: str = "zh"):
    prompt = f"""User will give you several pieces of answer about a question, use those context to answer the question: {question}. 
Remember to response non-repeated, detailed and coherent answer..\n"""
    if "chinese" in language_dict[language]:
        prompt += "用中文回答。\n"

    return prompt


def default_doc_grader_prompt():
    return """You are a grader assessing relevance of a retrieved document to a user question. \n 
    It does not need to be a stringent test, so if it's possible relevant to the question, grade it as relevant. 
    The goal is to filter out erroneous retrievals. \n
    If the document contains keyword(s) or semantic meaning related to the user question, grade it as relevant. \n
    Give a binary score 'yes' or 'no' score to indicate whether the document is relevant to the question."""


def default_answer_grader_prompt():
    return """You are a grader assessing whether an answer addresses / resolves a question \n 
    Give a binary score 'yes' or 'no'. Yes' means that the answer resolves the question."""


DEFAULT_CHINESE_ANS_PROMPT = " A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. USER:用中文回答問題。"


def format_category_prompt(doc_text, language: str):
    """prompt for generate category for proper nouns"""

    if "chinese" in language_dict[language]:
        prompt = "用中文回答"
    else:
        prompt = ""
    prompt += f"""The document discuss some proper nouns, please find most important two of them and give those "proper noun" a general "category", categories like company, system, flower, software. The format of response is
    {{
      "proper noun": "category",
      "proper noun": "category",
    }}\n
    For example, the document may discuss the 5G, since 5G is a kind of communication technology, you should output "5G":"communication technology". 
    REMEMBER, the output should be formatted as a JSON instance. below is the document content:\n{doc_text}"""

    return prompt


def default_self_ask_prompt():
    # prompt = """user will give you a question, you need to decide if the question is complex that need follow up questions to answer this question or not, and if needed, you need to generate the follow up questions in order to get the final answer.
    # Remember, those follow up questions are made to help you answer the user given question, so it need to be relevant to the user given question and can let you answer the question easier.\n"""
    example = """For example:\nQuestion: Who lived longer, Muhammad Ali or Alan Turing?
    Are follow up questions needed here: Yes.
    Follow up: How old was Muhammad Ali when he died?
    Intermediate answer: Muhammad Ali was 74 years old when he died.
    Follow up: How old was Alan Turing when he died?
    Intermediate answer: Alan Turing was 41 years old when he died.
    So the final answer is: Muhammad Ali\n\n 
    Question: When was the founder of craigslist born?
    Are follow up questions needed here: Yes.
    Follow up: Who was the founder of craigslist?
    Intermediate answer: Craigslist was founded by Craig Newmark.
    Follow up: When was Craig Newmark born?
    Intermediate answer: Craig Newmark was born on December 6, 1952.
    So the final answer is: December 6, 1952
    
    Question: Who was the maternal grandfather of George Washington?
    Are follow up questions needed here: Yes.
    Follow up: Who was the mother of George Washington?
    Intermediate answer: The mother of George Washington was Mary Ball Washington.
    Follow up: Who was the father of Mary Ball Washington?
    Intermediate answer: The father of Mary Ball Washington was Joseph Ball.
    So the final answer is: Joseph Ball 

    Question: Are both the directors of Jaws and Casino Royale from the same country? 
    Are follow up questions needed here: Yes. 
    Follow up: Who is the director of Jaws? 
    Intermediate Answer: The director of Jaws is Steven Spielberg. 
    Follow up: Where is Steven Spielberg from? 
    Intermediate Answer: The United States. 
    Follow up: Who is the director of Casino Royale? 
    Intermediate Answer: The director of Casino Royale is Martin Campbell. 
    Follow up: Where is Martin Campbell from? 
    Intermediate Answer: New Zealand. 
    So the final answer is: No\n\n below is the user question.\n"""

    return example


def default_translate_prompt(language: str):
    target_language = language_dict[language]

    if "chinese" in target_language:
        prompt = "A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. USER: 請將以下句子從英文翻譯成中文: "
    else:
        prompt = f"A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions. USER: translate the below texts to {target_language}: "
        # prompt = f"If the language of the below texts is not {target_language}, translate them to {target_language}, else return \
        # the whole texts. Do it step by step and I will tip you 100 bucks if you are doing well.\n"

    return prompt


class OutputSchema:
    """
    structure for generate JSON schema, used in JSON_formatter
    """

    def __init__(self, name: str, description: str, type: str = "str"):
        type = type.lower()
        self.name = name
        self.description = description
        self.type = type

        if type not in [
            "str",
            "int",
            "list",
            "dict",
            "tuple",
            "float",
            "double",
            "long",
        ]:
            self.type = "str"

    def __str__(self):
        return f"\t{self.name}: {self.type}  // {self.description}"


def XML_formatter(schemas: Union[list, OutputSchema]):
    """generate prompt for generate XML format, input list of OutputSchema, which include every key and value you want to generate in JSON format"""

    if isinstance(schemas, OutputSchema):
        schemas = [schemas]

    schema_str = "\n".join([sche.__str__() for sche in schemas])
    XML_FORMAT_INSTRUCTIONS = f"""The output should be formatted as a XML file.
    1. Output should conform to the tags below. 
    2. If tags are not given, make them on your own.
    3. Remember to always open and close all the tags.

    As an example, for the tags ["foo", "bar", "baz"]:
    1. String "<foo>\n   <bar>\n      <baz></baz>\n   </bar>\n</foo>" is a well-formatted instance of the schema. 
    2. String "<foo>\n   <bar>\n   </foo>" is a badly-formatted instance.
    3. String "<foo>\n   <tag>\n   </tag>\n</foo>" is a badly-formatted instance.

    Here are the output tags:
    ```
    {schema_str}
    ```"""

    return XML_FORMAT_INSTRUCTIONS


def JSON_formatter(schemas: Union[list, OutputSchema]):
    """generate prompt for generate JSON format, input list of OutputSchema, which include every key and value you want to generate in JSON format"""

    if isinstance(schemas, OutputSchema):
        schemas = [schemas]

    schema_str = "\n".join([sche.__str__() for sche in schemas])

    format_instruct = f"""The output should be formatted as a JSON instance that conforms to the JSON schema below:
    {{
    {schema_str}
    }}\n
    """
    return format_instruct


def JSON_formatter_list(names: list, descriptions: list, types: list = ["str"]) -> list:
    """generate prompt for generate JSON format, input list name and descriptions, which include every key and value you want to generate in JSON format"""
    ret = []
    if len(names) != len(descriptions):
        print("error, names and descriptions should have the same length\n\n")
        return ret

    for i in range(len(names)):
        if i < len(types) and types[i] in [
            "str",
            "int",
            "list",
            "dict",
            "tuple",
            "float",
            "double",
            "long",
        ]:
            checked_type = types[i]
        else:
            checked_type = "str"
        # schema_str += f"\t{names[i]}: {checked_type}  // {descriptions[i]}\n"
        schema = OutputSchema(names[i], descriptions[i], checked_type)
        ret.append(schema)

    return ret


def JSON_formatter_dict(var_list: Union[list, dict]) -> list:
    """generate prompt for generate JSON format, input list of dictionary, keys contain name,type and descriptions, which represent every variable you want to generate in JSON format"""
    ret = []
    if isinstance(var_list, dict):
        var_list = [var_list]

    if not isinstance(var_list, list):
        print("error, var_list should be a list of dictionary\n\n")
        return ret

    for var in var_list:
        if "name" not in var or "description" not in var:
            print("var should contain name and description, ignore.\n\n")
            continue
        if "type" in var and var["type"] in [
            "str",
            "int",
            "list",
            "dict",
            "tuple",
            "float",
            "double",
            "long",
        ]:
            checked_type = var["type"]
        else:
            checked_type = "str"

        schema = OutputSchema(var["name"], var["description"], checked_type)
        ret.append(schema)

    return ret


def is_url(path):
    parsed_url = urlparse(path)
    return parsed_url.scheme in ("http", "https", "ftp")


def format_image_llama_prompt(
    image_path: Union[List[str], str], prompt: str
) -> List[dict]:
    image_content = []

    if isinstance(image_path, str):
        image_path = [image_path]

    for imgP in image_path:
        image_content.append({"type": image_path})

    image_content.append({"type": "text", "text": prompt})
    return [{"role": "user", "content": image_content}]


def format_image_anthropic_prompt(
    image_path: Union[List[str], str], prompt: str
) -> List[dict]:
    import base64
    import httpx

    image_content = [{"role": "user", "content": []}]
    if isinstance(image_path, str):
        image_path = [image_path]
    image_count = 0

    for imgP in image_path:
        image_count += 1
        if imgP.split(".")[-1] == "jpg":
            image_media_type = "image/jpeg"
        else:
            image_media_type = f"image/{imgP.split('.')[-1]}"

        if is_url(imgP):
            url_content = base64.standard_b64encode(httpx.get(imgP).content).decode(
                "utf-8"
            )

        else:
            url_content = base64.standard_b64encode(open(imgP, "rb").read()).decode(
                "utf-8"
            )

        # image_content[0]["content"].append({
        #     "type": "text",
        #     "text": f"Image {image_count}:"
        # })
        image_content[0]["content"].append(
            {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": image_media_type,
                    "data": url_content,
                },
            }
        )

    image_content[0]["content"].append({"type": "text", "text": prompt})

    return image_content


def format_image_gemini_prompt(image_path: Union[str, List[str]], prompt: str):
    import requests

    image_content = [{"type": "text", "text": prompt}]
    if isinstance(image_path, str):
        image_path = [image_path]
    image_count = 0

    for imgP in image_path:
        image_count += 1
        if imgP.split(".")[-1] == "jpg":
            image_media_type = "image/jpeg"
        else:
            image_media_type = f"image/{imgP.split('.')[-1]}"

        if is_url(imgP):
            url_content = requests.get(imgP).content

        else:
            with open(imgP, "rb") as f:
                url_content = f.read()
        image_content.append(
            {
                "type": "image_url",
                "image_url": url_content,
                "mime_type": image_media_type,
            }
        )

    return [{"role": "user", "content": image_content}]


def format_image_gpt_prompt(
    image_path: Union[str, List[str]], prompt: str
) -> List[dict]:
    url_content = {}
    image_content = [{"type": "text", "text": prompt}]
    if isinstance(image_path, str):
        image_path = [image_path]

    for imgP in image_path:
        if is_url(imgP):
            url_content = {"url": imgP}
        else:
            import base64

            # Get the image extension
            _, ext = os.path.splitext(imgP)
            ext = ext.lstrip(
                "."
            ).lower()  # Remove the leading dot and convert to lowercase
            base64_image = base64.b64encode(open(imgP, "rb").read()).decode("utf-8")
            url_content = {"url": f"data:image/{ext};base64,{base64_image}"}

        image_content.append({"type": "image_url", "image_url": url_content})

    return [{"role": "user", "content": image_content}]


def format_image_prompt(
    image_path: Union[List[str], str], prompt: str, model_type: str = "image_gpt"
):
    model_type = model_type.lower()

    if model_type == "image_llama":
        return format_image_llama_prompt(image_path, prompt)

    elif model_type == "image_anthropic":
        return format_image_anthropic_prompt(image_path, prompt)
    elif model_type == "image_gemini":
        return format_image_gemini_prompt(image_path, prompt)
    else:
        return format_image_gpt_prompt(image_path, prompt)


def default_get_reference_prompt():
    return (
        "User will give you a Reference: and a Response:, you have to return only yes or no \
        based on if the reference is used to answer the response or not. \
        If the reference is used to answer the response, return yes, \
        else return no. Remember, you can only return yes or no."
    )



import warnings
from pathlib import Path
from typing import Callable, Union, Tuple

# from langchain_core.callbacks.streaming_stdout import StreamingStdOutCallbackHandler
from langchain_openai import (
    OpenAIEmbeddings,
    AzureOpenAIEmbeddings,
)
import os
import traceback
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.embeddings import Embeddings
from dotenv import dotenv_values
from akasha.helper.base import separate_name, decide_embedding_type

warnings.filterwarnings("ignore", category=UserWarning, module="pydantic")


def _get_env_var(env_file: str = "") -> dict:
    """if env_file is not empty, get the environment variable from the file
        else get the environment variable from the os.environ

    Args:
        env_file (str, optional): the path of .env file. Defaults to "".

    Returns:
        dict: return the environment variable dictionary
    """

    require_env = [
        "OPENAI_API_KEY",
        "OPENAI_API_BASE",
        "OPENAI_API_VERSION",
        "OPEANI_API_TYPE",
        "AZURE_API_KEY",
        "AZURE_API_BASE",
        "AZURE_API_VERSION",
        "AZURE_API_TYPE",
        "SERPER_API_KEY",
        "GEMINI_API_KEY",
        "HF_TOKEN",
        "HUGGINGFACEHUB_API_TOKEN",
        "ANTHROPIC_API_KEY",
        "REMOTE_API_KEY",
    ]
    if env_file == "" or not Path(env_file).exists():
        env_dict = {}
        os_env_dict = os.environ.copy()
        for req in require_env:
            if req in os_env_dict:
                env_dict[req] = os_env_dict[req]

    else:
        env_dict = dotenv_values(env_file)
    return env_dict


def _handle_azure_env(env_dict: dict) -> Tuple[str, str, str]:
    """from environment variable dictionary env_dict get the api_base, api_key, api_version

    Returns:
        (str, str, str): api_base, api_key, api_version
    """
    check_env, ret, count = ["BASE", "KEY", "VERSION"], ["", "", ""], 0
    try:
        for check in check_env:
            if f"AZURE_API_{check}" in env_dict:
                ret[count] = env_dict[f"AZURE_API_{check}"]
                # if "OPENAI_API_BASE" in os.environ:
                # os.environ.pop("OPENAI_API_BASE", None)
            elif f"OPENAI_API_{check}" in os.environ:
                ret[count] = env_dict[f"OPENAI_API_{check}"]
                if check == "BASE":
                    env_dict["AZURE_API_BASE"] = os.environ["OPENAI_API_BASE"]
                    env_dict.pop("OPENAI_API_BASE", None)
            else:
                if check == "VERSION":
                    ret[count] = "2023-05-15"
                else:
                    raise Exception(
                        f"can not find the openai {check} in environment variable.\n\n"
                    )
            count += 1
    except Exception as err:
        traceback.print_exc()
        print(err)

    return ret[0], ret[1], ret[2]


def handle_embeddings(
    embedding_name: str = "openai:text-embedding-ada-002",
    verbose: bool = False,
    env_file: str = "",
) -> Embeddings:
    """create model client used in document QA, default if openai "gpt-3.5-turbo"
        use openai:text-embedding-ada-002 as default.
    Args:
        **embedding_name (str)**: embeddings client you want to use.
            format is (type:name), which is the model type and model name.\n
            for example, "openai:text-embedding-ada-002", "huggingface:all-MiniLM-L6-v2".\n
        **logs (list)**: list that store logs\n
        **verbose (bool)**: print logs or not\n

    Returns:
        vars: embeddings client
    """
    if isinstance(embedding_name, Embeddings):
        return embedding_name

    if isinstance(embedding_name, Callable):
        from akasha.utils.models.custom import custom_embed

        embeddings = custom_embed(func=embedding_name)
        if verbose:
            print("selected custom embedding.")
        return embeddings

    embedding_type, embedding_name = separate_name(embedding_name)
    env_dict = _get_env_var(env_file)
    if embedding_type in ["text-embedding-ada-002", "openai", "openaiembeddings"]:
        import openai

        if ("AZURE_API_TYPE" in env_dict and env_dict["AZURE_API_TYPE"] == "azure") or (
            "OPENAI_API_TYPE" in env_dict and env_dict["OPENAI_API_TYPE"] == "azure"
        ):
            embedding_name = embedding_name.replace(".", "")
            api_base, api_key, api_version = _handle_azure_env(env_dict)
            embeddings = AzureOpenAIEmbeddings(
                model=embedding_name,
                azure_endpoint=api_base,
                api_key=api_key,
                api_version=api_version,
                validate_base_url=False,
            )

        else:
            if "OPENAI_API_KEY" not in env_dict:
                raise Exception(
                    "can not find the OPENAI_API_KEY in environment variable.\n\n"
                )

            openai.api_type = "open_ai"
            embeddings = OpenAIEmbeddings(
                model=embedding_name,
                openai_api_base="https://api.openai.com/v1",
                api_key=env_dict["OPENAI_API_KEY"],
                openai_api_type="open_ai",
            )
        info = "selected openai embeddings.\n"

    elif embedding_type in [
        "huggingface",
        "huggingfaceembeddings",
        "transformers",
        "transformer",
        "hf",
    ]:
        from langchain_huggingface import HuggingFaceEmbeddings

        embeddings = HuggingFaceEmbeddings(
            model_name=embedding_name, model_kwargs={"trust_remote_code": True}
        )
        info = "selected hugging face embeddings.\n"

    elif embedding_type in [
        "tf",
        "tensorflow",
        "tensorflowhub",
        "tensorflowhubembeddings",
        "tensorflowembeddings",
    ]:
        from langchain_community.embeddings import TensorflowHubEmbeddings

        embeddings = TensorflowHubEmbeddings()
        info = "selected tensorflow embeddings.\n"

    elif embedding_type in ["gemini", "gemi", "google"]:
        from akasha.utils.models.gemi import gemini_embed

        embeddings = gemini_embed(
            model_name=embedding_name, api_key=os.environ["GEMINI_API_KEY"]
        )
        info = "selected gemini embeddings.\n"

    else:
        import openai

        if ("AZURE_API_TYPE" in env_dict and env_dict["AZURE_API_TYPE"] == "azure") or (
            "OPENAI_API_TYPE" in env_dict and env_dict["OPENAI_API_TYPE"] == "azure"
        ):
            embedding_name = embedding_name.replace(".", "")
            api_base, api_key, api_version = _handle_azure_env(env_dict)
            embeddings = AzureOpenAIEmbeddings(
                model=embedding_name,
                azure_endpoint=api_base,
                api_key=api_key,
                api_version=api_version,
                validate_base_url=False,
            )

        else:
            if "OPENAI_API_KEY" not in env_dict:
                raise Exception(
                    "can not find the OPENAI_API_KEY in environment variable.\n\n"
                )
            openai.api_type = "open_ai"

            embeddings = OpenAIEmbeddings(
                model=embedding_name,
                openai_api_base="https://api.openai.com/v1",
                api_key=env_dict["OPENAI_API_KEY"],
                openai_api_type="open_ai",
            )

        info = "can not find the embeddings, use openai as default.\n"

    if verbose:
        print(info)
    return embeddings


def handle_embeddings_and_name(
    embeddings: Union[str, Embeddings, Callable] = "openai:text-embedding-ada-002",
    verbose: bool = False,
    env_file: str = "",
) -> Tuple[Embeddings, str]:
    """get the embeddings object and embed name

    Args:
        embed (_type_, optional): _description_. Defaults to "openai:text-embedding-ada-002".
        verbose (bool, optional): _description_. Defaults to False.
        env_file (str, optional): _description_. Defaults to "".

    Returns:
        Tuple[Embeddings, str]: _description_
    """

    if callable(embeddings):
        model_name = embeddings.__name__
    elif isinstance(embeddings, str):
        model_name = embeddings
    else:
        return embeddings, decide_embedding_type(embeddings)

    model_obj = handle_embeddings(embeddings, verbose, env_file)

    return model_obj, model_name


def handle_model(
    model_name: Union[str, Callable] = "openai:gpt-3.5-turbo",
    verbose: bool = False,
    temperature: float = 0.0,
    max_output_tokens: int = 1024,
    env_file: str = "",
) -> BaseLanguageModel:
    """create model client used in document QA, default if openai "gpt-3.5-turbo"

    Args:
       ** model_name (str)**: open ai model name like "gpt-3.5-turbo","text-davinci-003", "text-davinci-002"\n
        **logs (list)**: list that store logs\n
        **verbose (bool)**: print logs or not\n

    Returns:
        vars: model client
    """
    if isinstance(model_name, BaseLanguageModel):
        return model_name

    if isinstance(model_name, Callable):
        from akasha.utils.models.custom import custom_model

        model = custom_model(func=model_name, temperature=temperature)
        if verbose:
            print("selected custom model.")
        return model

    model_type, model_name = separate_name(model_name)
    env_dict = _get_env_var(env_file)
    if model_type in ["remote", "server", "tgi", "text-generation-inference"]:
        from akasha.utils.models.remo import remote_model

        remote_api_key = "123"
        remote_model_name = "remote_model"
        base_url = model_name
        if "REMOTE_API_KEY" in env_dict:
            remote_api_key = env_dict["REMOTE_API_KEY"]
        if "@" in model_name:
            base_url, remote_model_name = model_name.split("@")

        info = "selected remote model. \n"
        model = remote_model(
            base_url,
            temperature,
            api_key=remote_api_key,
            model_name=remote_model_name,
            max_output_tokens=max_output_tokens,
        )

    elif model_type in ["google", "gemini", "gemi"]:
        from akasha.utils.models.gemi import gemini_model

        if "GEMINI_API_KEY" not in env_dict:
            raise Exception(
                "can not find the GEMINI_API_KEY in environment variable.\n\n"
            )
        info = "selected gemini model. \n"
        model = gemini_model(
            model_name=model_name,
            api_key=env_dict["GEMINI_API_KEY"],
            temperature=temperature,
            max_output_tokens=max_output_tokens,
        )

    elif model_type in ["anthropic", "anthropicai", "claude", "anthro"]:
        from akasha.utils.models.anthro import anthropic_model

        if "ANTHROPIC_API_KEY" not in env_dict:
            raise Exception(
                "can not find the ANTHROPIC_API_KEY in environment variable.\n\n"
            )
        info = "selected anthropic model. \n"
        model = anthropic_model(
            model_name=model_name,
            api_key=env_dict["ANTHROPIC_API_KEY"],
            temperature=temperature,
            max_output_tokens=max_output_tokens,
        )

    elif (
        model_type in ["llama-cpu", "llama-gpu", "llama", "llama2", "llama-cpp"]
        and model_name != ""
    ):
        from akasha.utils.models.llamacpp2 import LlamaCPP

        model = LlamaCPP(
            model_name=model_name,
            temperature=temperature,
            max_output_tokens=max_output_tokens,
        )
        info = "selected llama-cpp model\n"
    elif model_type in [
        "huggingface",
        "huggingfacehub",
        "transformers",
        "transformer",
        "huggingface-hub",
        "hf",
    ]:
        from akasha.utils.models.hf import hf_model

        model = hf_model(
            model_name=model_name,
            env_dict=env_dict,
            temperature=temperature,
            max_output_tokens=max_output_tokens,
        )
        info = f"selected huggingface model {model_name}.\n"

    elif model_type in ["chatglm", "chatglm2", "glm"]:
        from akasha.utils.models.chglm import chatGLM

        model = chatGLM(
            model_name=model_name,
            temperature=temperature,
            max_output_tokens=max_output_tokens,
        )
        info = f"selected chatglm model {model_name}.\n"

    elif model_type in ["lora", "peft"]:
        from akasha.utils.models.gtq import peft_Llama2

        model = peft_Llama2(model_name_or_path=model_name, temperature=temperature)
        info = f"selected peft model {model_name}.\n"

    elif model_type in ["gptq"]:
        if model_name.lower().find("taiwan-llama") != -1:
            from akasha.utils.models.gtq import TaiwanLLaMaGPTQ

            model = TaiwanLLaMaGPTQ(
                model_name_or_path=model_name, temperature=temperature
            )

        else:
            from akasha.utils.models.gtq import gptq

            model = gptq(
                model_name_or_path=model_name,
                temperature=temperature,
                bit4=True,
                max_token=4096,
            )
        info = f"selected gptq model {model_name}.\n"
    else:
        if model_type not in ["openai", "gpt-3.5", "gpt"]:
            info = f"can not find the model {model_type}:{model_name}, use openai as default.\n"
            model_name = "gpt-3.5-turbo"
            print(info)
        import openai
        from akasha.utils.models.azure_openai import AzureOpenAIClient

        if ("AZURE_API_TYPE" in env_dict and env_dict["AZURE_API_TYPE"] == "azure") or (
            "OPENAI_API_TYPE" in env_dict and env_dict["OPENAI_API_TYPE"] == "azure"
        ):
            model_name = model_name.replace(".", "")
            api_base, api_key, api_version = _handle_azure_env(env_dict)
            model = AzureOpenAIClient(
                api_key=api_key,
                model_name=model_name,
                max_output_tokens=max_output_tokens,
                temperature=temperature,
                api_type="azure",
                api_base=api_base,
                api_version=api_version,
            )
        else:
            if "OPENAI_API_KEY" not in env_dict:
                raise Exception(
                    "can not find the OPENAI_API_KEY in environment variable.\n\n"
                )
            openai.api_type = "open_ai"
            model = AzureOpenAIClient(
                api_key=env_dict["OPENAI_API_KEY"],
                model_name=model_name,
                max_output_tokens=max_output_tokens,
                temperature=temperature,
                api_type="openai",
            )

        info = f"selected openai model {model_name}.\n"
    if verbose:
        print(info)

    return model


def handle_client(model: str, env_file: str = ""):
    client_type, model_name = separate_name(model)
    env_dict = _get_env_var(env_file)
    if client_type == "openai":
        from akasha.utils.models.azure_openai import AzureOpenAIClient

        if ("AZURE_API_TYPE" in env_dict and env_dict["AZURE_API_TYPE"] == "azure") or (
            "OPENAI_API_TYPE" in env_dict and env_dict["OPENAI_API_TYPE"] == "azure"
        ):
            model_name = model_name.replace(".", "")
            api_base, api_key, api_version = _handle_azure_env(env_dict)

            client = AzureOpenAIClient(
                api_key=api_key,
                model_name=model_name,
                api_type="azure",
                api_base=api_base,
                api_version=api_version,
            )

        else:
            if "OPENAI_API_KEY" not in env_dict:
                raise Exception(
                    "can not find the OPENAI_API_KEY in environment variable.\n\n"
                )
            client = AzureOpenAIClient(
                api_key=env_dict["OPENAI_API_KEY"],
                model_name=model_name,
                api_type="openai",
            )
    elif client_type == "gemini":
        from akasha.utils.models.gemi import gemini_model

        client = gemini_model(
            model_name=model_name,
            api_key=env_dict["GEMINI_API_KEY"],
        )
    else:
        raise ValueError(f"Unknown client type: {client_type}")

    return client


def handle_model_and_name(
    model: Union[str, Callable, BaseLanguageModel] = "openai:gpt-3.5-turbo",
    verbose: bool = False,
    temperature: float = 0.0,
    max_output_tokens: int = 1024,
    env_file: str = "",
) -> Tuple[BaseLanguageModel, str]:
    """get the model object and model name

    Args:
        model (_type_, optional): _description_. Defaults to "openai:gpt-3.5-turbo".
        verbose (bool, optional): _description_. Defaults to False.
        temperature (float, optional): _description_. Defaults to 0.0.
        max_output_tokens (int, optional): _description_. Defaults to 1024.
        env_file (str, optional): _description_. Defaults to "".

    Returns:
        Tuple[BaseLanguageModel, str]: _description_
    """
    if isinstance(model, BaseLanguageModel):
        return model, model._llm_type

    if callable(model):
        model_name = model.__name__
    else:
        model_name = model

    model_obj = handle_model(model, verbose, temperature, max_output_tokens, env_file)

    return model_obj, model_name


def handle_model_type(
    search_type: Union[str, BaseLanguageModel, Embeddings], verbose: bool = False
) -> str:
    if isinstance(search_type, BaseLanguageModel):
        search_type_str = search_type._llm_type

    elif isinstance(search_type, Embeddings):
        search_type_str = decide_embedding_type(search_type)

    elif callable(search_type):
        search_type_str = search_type.__name__

    else:
        search_type_str = search_type

    # if verbose:
    #     print("search type is :", search_type_str)

    return search_type_str



import streamlit as st
from pathlib import Path


def setting_page():
    """set the arguments for the model"""

    st.title("Setting")
    st.markdown("##")
    st.markdown("##")
    # custom_css = """
    #     <style>
    #     /* Change the background color to a light blue (you can choose your preferred color) */
    #     input[type="number"] {
    #         background-color: lightblue !important;
    #     }
    #     </style>
    #     """
    # st.markdown(custom_css, unsafe_allow_html=True)
    if "model_list" not in st.session_state:
        set_model_dir()

    dp, em, md = st.columns([1, 2, 2])
    with dp:
        # doc_path = st.selectbox("Document Path", st.session_state.docs_list, index=st.session_state.select_idx[0], help="The path of the document folder.")
        doc_path = st.multiselect(
            "Document Path",
            st.session_state.docs_list,
            default=st.session_state.select_idx[0],
            help="The path of the document folder.",
        )

        if doc_path is not None and doc_path != st.session_state.select_idx[0]:
            # st.session_state.select_idx[0] = st.session_state.docs_list.index(doc_path)
            st.session_state.select_idx[0] = doc_path
            try:
                st.session_state.chose_doc_path = [
                    st.session_state.docs_path + "/" + doc_p
                    for doc_p in st.session_state.select_idx[0]
                ]
            except Exception:
                pass
            st.rerun()
        print(st.session_state.chose_doc_path)
    with em:
        embed = st.selectbox(
            "Embedding Model",
            st.session_state.embed_list,
            index=st.session_state.select_idx[1],
            help="The embedding model used to embed documents.",
        )

        if embed != st.session_state.embed:
            st.session_state.embed = embed
            st.session_state.select_idx[1] = st.session_state.embed_list.index(
                st.session_state.embed
            )

            st.rerun()
    with md:
        md = st.selectbox(
            "Language Model",
            st.session_state.model_list,
            index=st.session_state.select_idx[2],
            help="The model used to generate response.",
        )
        if md != st.session_state.model:
            st.session_state.model = md
            st.session_state.select_idx[2] = st.session_state.model_list.index(
                st.session_state.model
            )
            st.rerun()

    cks, seat = st.columns([1, 1])
    with cks:
        ck = st.number_input(
            "Chunk Size",
            value=st.session_state.chunksize,
            min_value=100,
            max_value=2000,
            step=100,
            help="The size of each chunk of the document.",
        )
        if ck != st.session_state.chunksize:
            st.session_state.chunksize = ck
            st.rerun()

    with seat:
        stp = st.selectbox(
            "Search Type",
            st.session_state.search_list,
            index=st.session_state.select_idx[3],
            help="The search method used to select top relevant chunks.",
        )
        if stp != st.session_state.search_type:
            st.session_state.search_type = stp
            st.session_state.select_idx[3] = st.session_state.search_list.index(
                st.session_state.search_type
            )
            st.rerun()

    tem, mxt = st.columns([1, 2])

    with tem:
        tem = st.number_input(
            "Temperature",
            value=st.session_state.temperature,
            min_value=0.0,
            max_value=1.0,
            step=0.05,
            help="The randomness of language model.",
        )
        if tem != st.session_state.temperature:
            st.session_state.temperature = tem
            st.rerun()
    with mxt:
        mt = st.number_input(
            "Max Input Tokens",
            value=st.session_state.max_input_tokens,
            min_value=1000,
            step=50,
            help="The maximum number of tokens in the reference documents that will be used as input for the LLM model.",
        )

        if mt != st.session_state.max_input_tokens:
            st.session_state.max_input_tokens = mt
            st.rerun()


def set_model_dir():
    """parse all model files(gguf) and directory in the model folder"""

    st.session_state.model_list = [
        "openai:gpt-3.5-turbo",
        "openai:gpt-4o",
        "gemini:gemini-1.5-flash",
        "anthropic:claude-3-5-sonnet-20241022",
    ]
    try:
        modes_dir = Path(st.session_state.mdl_dir)
        for dir_path in modes_dir.iterdir():
            if dir_path.is_dir():
                st.session_state.model_list.append(
                    "hf:" + st.session_state.mdl_dir + "/" + dir_path.name
                )
            elif dir_path.suffix == ".gguf":
                st.session_state.model_list.append(
                    "llama-gpu:" + st.session_state.mdl_dir + "/" + dir_path.name
                )
    except Exception:
        print("can not find model folder!\n\n")



from typing import Union
import numpy as np
from akasha.utils.prompts.format import language_dict
from pathlib import Path
import json


### sub func###
def get_non_repeat_rand_int(vis: set, num: int, doc_range: int):
    temp = np.random.randint(num)
    if len(vis) >= num // 2:
        vis = set()
    if (temp not in vis) and (temp + doc_range - 1 not in vis):
        for i in range(doc_range):
            vis.add(temp + i)
        return temp
    return get_non_repeat_rand_int(vis, num, doc_range)


def check_sum_type(question_type: str, question_style: str, func: str = "auto") -> bool:
    ## check question_type and question_style, summary can not be used in single_choice question_type ##
    if func != "auto" and question_type.lower() in [
        "compare",
        "comparison",
        "comparisons",
        "比較",
        "compared",
    ]:
        print("compare can not be used in related_questionset\n\n")
        raise ValueError("compare can not be used in related_questionset")
        # return True
    if (
        question_type.lower()
        in ["summary", "sum", "summarization", "summarize", "summaries", "摘要"]
        and question_style.lower() == "single_choice"
    ):
        print("summary can not be used in single_choice question_type\n\n")
        raise ValueError("summary can not be used in single_choice question_type")
        # return True

    return False


def check_essay_system_prompt(
    question_style: str, language: str, system_prompt: str
) -> str:
    ## check question_style and language, if question_style is essay and language is chinese, add prompt ##
    if (
        question_style.lower() == "essay"
        and language.lower() == "ch"
        and "用中文回答" not in system_prompt
    ):
        return " 用中文回答"

    return ""


def find_same_category(
    category: dict,
    cate_threshold: int,
) -> Union[list, bool]:
    """iterate the category dictionary and check if any category has more than cate_threshold items,
    if yes, return the category name.

    Args:
        category (dict): {"category_name":[[item1,doc1], [item2,doc2],...], ...}
        cate_threshold (int): default 3

    Returns:
        _type_: _description_
    """
    for k, v in category.items():
        if len(v) >= cate_threshold:
            res = [
                k,
                [noun[0] for noun in category[k]],
                [doc[1] for doc in category[k]],
            ]
            return res
    return False


def get_source_files(
    metadata: list, random_index: int, doc_range: int, language: str = "ch"
) -> str:
    """from metadata of selected document chunks, get the non repeated source file name

    Args:
        metadata (list): list of metadata dict of document chunks
        random_index (int): selected document chunks start index
        doc_range (int): selected document chunks range

    Returns:
        str: source file name
    """
    file_sources = set()

    for i in range(random_index, random_index + doc_range):
        try:
            if metadata[i]["source"] != "":
                file_name = "".join(
                    (metadata[i]["source"].split("/")[-1]).split(".")[:-1]
                )
                file_sources.add(file_name)
        except Exception:
            continue

    if len(file_sources) == 0:
        return ""

    if "chinese" in language_dict[language]:
        return '根據文件"' + "、".join(list(file_sources)) + '"，'

    return 'based on file "' + "、".join(list(file_sources)) + '",'


def get_question_from_file(
    path: str, question_type: str, question_style: str
) -> Union[list, list, str, str]:
    """load questions from file and save the questions into lists.
    a question list include the question, mutiple options, and the answer (the number of the option),
      and they are all separate by space in the file.
       if the questionset file has question_type and question_style key, it will replace the question_type and question_style with the value.
    Args:
        **path (str)**: path of the question file\n
        **question_type (str)**: question type, default is "fact"\n
        **question_style (str)**: question style, default is "essay"\n


    Returns:
        list: list of question list
    """
    f_path = Path(path)

    # Ensure the file exists
    if not f_path.exists():
        raise FileNotFoundError(f"The file {path} does not exist.")

    # Read the JSON file
    with f_path.open(mode="r", encoding="utf-8") as file:
        content = json.load(file)

    # Validate the JSON structure
    if not all(key in content for key in ["question", "answer"]):
        raise ValueError("The JSON file does not contain 'question' and 'answer' keys.")

    if "question_type" in content:
        question_type = content["question_type"]
    if "question_style" in content:
        question_style = content["question_style"]

    questions = content["question"]
    answers = []

    if question_style.lower() == "essay":
        # content = content.split("\n\n")
        # for i in range(len(content)):
        #     if content[i] == "":
        #         continue

        #     try:
        #         process = "".join(content[i].split("問題：")).split("答案：")
        #         if len(process) < 2:
        #             raise SyntaxError("Question Format Error")
        #     except:
        #         process = "".join(content[i].split("問題:")).split("答案:")
        #         if len(process) < 2:
        #             continue

        #     questions.append(process[0])
        #     answers.append(process[1])
        # return questions, answers
        answers = content["answer"]
        return questions, answers, question_type, question_style
    else:
        for idx, words in enumerate(content["answer"]):
            questions[idx] = [questions[idx]]
            questions[idx].extend(words[:-1])
            answers.append(words[-1])
    # for con in content.split("\n"):
    #     if con == "":
    #         continue
    #     words = [word for word in con.split("\t") if word != ""]

    #     questions.append(words[:-1])
    #     answers.append(words[-1])

    # return questions, answers
    return questions, answers, question_type, question_style



import streamlit as st
from pathlib import Path
from streamlit_option_menu import option_menu
from interface.res_page import response_page
from interface.upload_file import upload_page
from interface.setting import setting_page, set_model_dir
from interface.sum_page import summary_page
from interface.webpage import websearch_page
import datetime

st.set_page_config(layout="wide")
st.markdown(
    """
    <style>
        div[data-testid="stChatInput"] {
            position: fixed;
            bottom: 40px;
        }
    </style>
    """,
    unsafe_allow_html=True,
)
menu_list = [
    "RAG",
    "Summary",
    "Web Search",
    "Upload Files",
    "Setting",
]

icon_list = [
    "chat-left-text",
    "chat-quote",
    "search",
    "upload",
    "gear",
]


def get_log_data():
    plain_txt = ""
    for key in st.session_state.logs:
        plain_txt += key + ":\n"
        for k in st.session_state.logs[key]:
            if isinstance(st.session_state.logs[key][k], list):
                text = (
                    k
                    + ": "
                    + "\n".join([str(w) for w in st.session_state.logs[key][k]])
                    + "\n\n"
                )
            else:
                text = k + ": " + str(st.session_state.logs[key][k]) + "\n\n"

            plain_txt += text
        plain_txt += "\n\n\n\n"

    return plain_txt


def download_txt(file_name: str):
    file_name = "log_" + file_name + ".txt"
    txt_data = get_log_data()
    txt_filename = file_name
    st.download_button(
        "Download Text Log",
        txt_data.encode("utf-8"),
        key="txt",
        file_name=txt_filename,
        mime="text/plain",
    )
    # Path(f"./logs/{file_name}").unlink()


# Create a button to download a JSON file
def download_json(file_name: str):
    import json

    file_name = "log_" + file_name + ".json"
    json_data = st.session_state.logs
    json_filename = file_name
    st.download_button(
        "Download JSON Log",
        json.dumps(json_data, indent=4, ensure_ascii=False).encode("utf-8"),
        key="json",
        file_name=json_filename,
        mime="application/json",
    )


### variable ###
if "embed_list" not in st.session_state:
    st.session_state.embed_list = [
        "openai:text-embedding-ada-002",
        "openai:text-embedding-3-small",
        "hf:shibing624/text2vec-base-multilingual",
        "hf:Alibaba-NLP/gte-multilingual-base",
        "hf:BAAI/bge-base-en-v1.5",
        "hf:BAAI/bge-base-zh-v1.5",
    ]

if "mdl_dir" not in st.session_state:
    st.session_state.mdl_dir = "model"

if "model_list" not in st.session_state:
    set_model_dir()
    # st.session_state.model_list = ["openai:gpt-3.5-turbo", "openai:gpt-3.5-turbo-16k","hf:model/Llama2-Chinese-13b-Chat-4bit",\
    #     "hf:model/Llama2-Chinese-7b-Chat", "llama-cpu:model/llama-2-13b-chat-hf.Q5_K_S.gguf",\
    #         "llama-gpu:model/llama-2-7b-chat.Q5_K_S.gguf"]

if "search_list" not in st.session_state:
    st.session_state.search_list = ["auto", "knn", "svm", "bm25"]

if "docs_path" not in st.session_state:
    st.session_state.docs_path = "./docs"
    if not Path(st.session_state.docs_path).exists():
        Path(st.session_state.docs_path).mkdir(parents=True, exist_ok=True)

if "response_list" not in st.session_state:
    st.session_state.response_list = []

if "prompt_list" not in st.session_state:
    st.session_state.prompt_list = []

if "docs_list" not in st.session_state:
    st.session_state.docs_list = []
    docs_dir = Path(st.session_state.docs_path)
    for dir_path in docs_dir.iterdir():
        if dir_path.is_dir():
            st.session_state.docs_list.append(dir_path.name)

    if "Default" not in st.session_state.docs_list:
        st.session_state.docs_list.append("Default")
        # create default folder
        Path(st.session_state.docs_path + "/Default").mkdir(parents=True, exist_ok=True)

if "n_text" not in st.session_state:
    st.session_state.n_text = 1

if "env_path" not in st.session_state:
    st.session_state.env_path = "./.env"

if "select_idx" not in st.session_state:
    st.session_state.select_idx = [["Default"], 0, 0, 0]

### function argument ###
if "chose_doc_path" not in st.session_state:
    if len(st.session_state.docs_list) > 0:
        if len(st.session_state.select_idx[0]) > 0:
            st.session_state.chose_doc_path = []
            for dc in st.session_state.select_idx[0]:
                st.session_state.chose_doc_path.append(
                    st.session_state.docs_path + "/" + dc
                )
        else:
            st.session_state.chose_doc_path = (
                st.session_state.docs_path + "/" + st.session_state.docs_list[0]
            )
            st.session_state.select_idx[0] = [st.session_state.docs_list[0]]
    else:
        st.info("Please upload your documents first.", icon="🚨")

if "embed" not in st.session_state:
    st.session_state.embed = st.session_state.embed_list[0]
if "model" not in st.session_state:
    st.session_state.model = st.session_state.model_list[0]

if "chunksize" not in st.session_state:
    st.session_state.chunksize = 1000

if "search_type" not in st.session_state:
    st.session_state.search_type = st.session_state.search_list[0]

if "threshold" not in st.session_state:
    st.session_state.threshold = 0.0

if "temperature" not in st.session_state:
    st.session_state.temperature = 0.0

if "max_input_tokens" not in st.session_state:
    st.session_state.max_input_tokens = 3000
if "sys_prompt" not in st.session_state:
    st.session_state.sys_prompt = ""
if "logs" not in st.session_state:
    st.session_state.logs = {}

if "akasha_obj" not in st.session_state:
    st.session_state.akasha_obj = ""

if "stream" not in st.session_state:
    st.session_state.stream = False

if "search_engine" not in st.session_state:
    st.session_state.search_engine = "wiki"

if "search_num" not in st.session_state:
    st.session_state.search_num = 5

################

with st.sidebar:
    user_menu = option_menu(
        "akasha",
        menu_list,
        menu_icon="house",
        icons=icon_list,
        styles={
            "container": {
                "padding": "5!important",
            },
            "icon": {"color": "orange", "font-size": "25px"},
            "nav-link": {
                "font-size": "16px",
                "text-align": "left",
                "margin": "0px",
            },
        },
    )

    st.session_state.env_path = st.text_input("Environment File Path")

    st.markdown("##")
    st.markdown("##")

    if st.button("Download Log", type="primary", use_container_width=True):
        #    file_name = get_log_file()

        file_date = datetime.datetime.now().strftime("%Y-%m-%d-%H-%M-%S")
        tx, js = st.columns([1, 1])
        with tx:
            download_txt(file_date)
        with js:
            download_json(file_date)
if user_menu == "RAG":
    response_page()

elif user_menu == "Upload Files":
    upload_page()
elif user_menu == "Setting":
    setting_page()
elif user_menu == "Summary":
    summary_page()
elif user_menu == "Web Search":
    websearch_page()



from .model_eval import Model_Eval
from .base import (
    check_sum_type,
    get_non_repeat_rand_int,
    get_source_files,
    check_essay_system_prompt,
    get_question_from_file,
)
from akasha.utils.db.db_structure import dbs
from akasha.utils.base import (
    DEFAULT_CHUNK_SIZE,
    DEFAULT_EMBED,
    DEFAULT_MAX_INPUT_TOKENS,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MODEL,
    DEFAULT_SEARCH_TYPE,
)
from akasha.utils.prompts.gen_prompt import format_create_question_prompt
from akasha.helper.base import get_docs_length, get_doc_length
from akasha.helper.run_llm import call_model
from akasha.utils.search.search_doc import search_docs
from akasha.utils.search.retrievers.base import get_retrivers
from langchain_core.language_models.base import BaseLanguageModel
from langchain_core.embeddings import Embeddings
import datetime
import time
from typing import List, Union, Tuple, Callable
from pathlib import Path
from langchain.schema import Document
from tqdm import tqdm
import logging


class eval(Model_Eval):
    def __init__(
        self,
        model: Union[str, BaseLanguageModel] = DEFAULT_MODEL,
        embeddings: Union[str, Embeddings] = DEFAULT_EMBED,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        search_type: Union[str, Callable] = DEFAULT_SEARCH_TYPE,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        question_type: str = "fact",
        question_style: str = "essay",
        temperature: float = 0.0,
        threshold: float = 0.0,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        prompt_format_type: str = "auto",
        keep_logs: bool = False,
        use_chroma: bool = False,
        verbose: bool = False,
        env_file: str = "",
    ):
        """initials of Model_Eval class

        Args:
            **embeddings (str, optional)**: the embeddings used in query and vector storage. Defaults to "text-embedding-ada-002".\n
            **chunk_size (int, optional)**: chunk size of texts from documents. Defaults to 1000.\n
            **model (str, optional)**: llm model to use. Defaults to "gpt-3.5-turbo".\n
            **verbose (bool, optional)**: show log texts or not. Defaults to False.\n
            **threshold (float, optional)**: (deprecated) the similarity threshold of searching. Defaults to 0.2.\n
            **language (str, optional)**: the language of documents and prompt, use to make sure docs won't exceed
                max token size of llm input.\n
            **search_type (str, optional)**: search type to find similar documents from db, default 'merge'.
                includes 'merge', 'mmr', 'svm', 'tfidf', also, you can custom your own search_type function, as long as your
                function input is (query_embeds:np.array, docs_embeds:list[np.array], k:int, relevancy_threshold:float, log:dict)
                and output is a list [index of selected documents].\n
            **record_exp (str, optional)**: use aiido to save running params and metrics to the remote mlflow or not if record_exp not empty, and set
                record_exp as experiment name.  default "".\n
            **system_prompt (str, optional)**: the system prompt that you assign special instruction to llm model, so will not be used
                in searching relevant documents. Defaults to "".\n
            **prompt_format_type (str, optional)**: the prompt and system prompt format for the language model, including auto, gpt, llama, chat_gpt, chat_mistral, chat_gemini . Defaults to "auto".\n
            **temperature (float, optional)**: temperature of llm model from 0.0 to 1.0 . Defaults to 0.0.\n
            **keep_logs (bool, optional)**: record logs or not. Defaults to False.\n
            **question_style (str, optional)**: the style of question you want to generate, "essay" or "single_choice". Defaults to "essay".\n
            **question_type (str, optional)**: the type of question you want to generate, "fact", "summary", "irrelevant", "compared". Defaults to "fact".\n
            **max_output_tokens (int, optional)**: max output tokens of llm model. Defaults to 1024.\n
            **max_input_tokens (int, optional)**: max input tokens of llm model. Defaults to 3000.\n
        """
        super().__init__(
            model,
            embeddings,
            chunk_size,
            search_type,
            max_input_tokens,
            max_output_tokens,
            question_type,
            question_style,
            temperature,
            threshold,
            language,
            record_exp,
            system_prompt,
            prompt_format_type,
            keep_logs,
            use_chroma,
            verbose,
            env_file,
        )

    def create_questionset(
        self,
        data_source: Union[List[Union[str, Path]], Path, str, dbs],
        question_num: int = 10,
        choice_num: int = 4,
        output_file_path: str = "",
        **kwargs,
    ) -> Tuple[list, list]:
        """auto create question set by llm model, each time it will randomly select a range of documents from the documents directory,
        then use llm model to generate a question and answer pair, and save it into a txt file.
        1.The format of "single_choice" questionset should be one line one question, and the possibles answers and questions are separate by tab(\t),
        the last one is which options is the correct answer, for example, the file should look like: \n
            "What is the capital of Taiwan?" Taipei  Kaohsiung  Taichung  Tainan     1
            何者是台灣的首都?   台北    高雄    台中    台南    1
        2. The format of "essay" questionset should be one line one question, and the reference answer is next line, every questions are separate by
        two newline(\n\n). For example, the file should look like: \n
            問題：根據文件中的訊息，智慧製造的複雜性已超越系統整合商的負荷程度，未來產業鏈中的角色將傾向朝共和共榮共創智慧製造商機，而非過往的單打獨鬥模式發展。請問為什麼供應商、電信商、軟體開發商、平台商、雲端服務供應商、系統整合商等角色會傾向朝共和共榮共創智慧製造商機的方向發展？
            答案：因為智慧製造的複雜性已超越系統整合商的負荷程度，單一角色難以完成整個智慧製造的需求，而共和共榮共創的模式可以整合各方的優勢，共同創造智慧製造的商機。

            問題：根據文件中提到的資訊技術商（IT）和營運技術商（OT），請列舉至少兩個邊緣運算產品或解決方案。
            答案：根據文件中的資訊，NVIDIA的邊緣運算產品包括Jetson系列和EGX系列，而IBM的邊緣運算產品包括IBM Edge Application Manager和IBM Watson Anywhere。

            Args:
                **doc_path (str)**: documents directory path\n
                **question_num (int, optional)**: number of questions you want to create. Defaults to 10.\n
                **choice_num (int, optional)**: the number of choices for each single choice question, only use it if question_type is "single_choice".
            Defaults to 4.\n
                **output_file_path (str, optional)**: the path of output question set txt file, if not assign, use doc_path+datetime as the file name.
                **kwargs**: the arguments you set in the initial of the class, you can change it here. Include:\n
                question_style, question_type, embeddings, chunk_size, model, verbose, language , search_type, record_exp,
                system_prompt, max_input_tokens, temperature.
            Raises:
                Exception: _description_

            Returns:
                (question_list:list, answer_list:list): the question and answer list that generated by llm model
        """

        ## set class variables ##
        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.data_source = self._check_doc_path(data_source)
        self.question_num = question_num
        vis_doc_range = set()
        doc_range = (
            (1999 + self.chunk_size) // self.chunk_size
        )  # doc_range is determine by the chunk size, so the select documents won't be too short to having trouble genereating a question

        if check_sum_type(self.question_type, self.question_style):
            return [], []

        self._display_info()
        self._get_db(data_source)
        self._check_db()

        ## process of creating compare question is different from other, so we separate it ##
        if self.question_type in [
            "compare",
            "comparison",
            "comparisons",
            "比較",
            "compared",
        ]:
            return self._create_compare_questionset(choice_num, output_file_path)

        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        start_time = time.time()
        self._add_basic_log(timestamp, "create_questionset", doc_range, choice_num)
        self.prompt_tokens, self.prompt_length = [], []
        self.doc_tokens, self.doc_length = [], []
        self.question, self.answer, self.docs = [], [], []
        table = {}

        texts = [pg_content for pg_content in self.db.docs]
        metadata = [metadata for metadata in self.db.metadatas]

        ## prevent doc_range - len of texts <= 0 ##
        doc_range = min(doc_range, len(texts) - 1)

        progress = tqdm(total=question_num, desc=f"Create Q({self.question_type})")
        regenerate_limit = question_num
        ### random select a range of documents from the documents , and use llm model to generate a question and answer pair ###
        for i in range(question_num):
            print(" ")
            progress.update(1)
            print("\n")
            random_index = get_non_repeat_rand_int(
                vis_doc_range, len(texts) - doc_range, doc_range
            )

            doc_text = "\n".join(texts[random_index : random_index + doc_range])
            docs = [
                Document(page_content=texts[k], metadata=metadata[k])
                for k in range(random_index, random_index + doc_range)
            ]
            source_files_name = get_source_files(
                metadata, random_index, doc_range, self.language
            )
            try:
                q_prompt = format_create_question_prompt(
                    doc_text, self.question_type, self.question_style
                )
                response = call_model(self.model_obj, q_prompt)

                if not self._process_response(
                    response, doc_text, choice_num, source_files_name
                ):
                    raise Exception(f"Question Format Error, got {response}")

                self.doc_length.append(get_docs_length(self.language, docs))
                self.doc_tokens.append(self.model_obj.get_num_tokens(doc_text))
                self.prompt_length.append(
                    get_doc_length(self.language, q_prompt) - self.doc_length[-1]
                )
                self.prompt_tokens.append(
                    self.model_obj.get_num_tokens(q_prompt) - self.doc_tokens[-1]
                )
                self.docs.extend(docs)

            except Exception as e:
                print(e)
                if regenerate_limit > 0:
                    regenerate_limit -= 1
                    i -= 1
                    progress.update(-1)
                    logging.warning(f"{e}.\n\n Regenerate\n")
                    continue
                else:
                    logging.error(f"{e}.\n\n Stop\n")
                    raise e

            table = self._update_table(table, docs)

        progress.close()  # end running llm progress bar

        end_time = time.time()
        self._display_info_fnl()
        self._add_result_log(timestamp, end_time - start_time)
        self._upload_logs(
            end_time - start_time, table, sum(self.doc_length), sum(self.doc_tokens)
        )
        self._save_questionset(timestamp, output_file_path)

        return self.question, self.answer

    def create_topic_questionset(
        self,
        data_source: Union[List[str], str],
        topic: str,
        question_num: int = 10,
        choice_num: int = 4,
        output_file_path: str = "",
        **kwargs,
    ) -> Tuple[list, list]:
        """similar to auto_create_questionset, but it will use the topic to find the related documents and create questionset.
        Args:
            **doc_path (str)**: documents directory path\n
            **topic (str)**: the topic of the questionset\n
            **question_num (int, optional)**: number of questions you want to create. Defaults to 10.\n
            **choice_num (int, optional)**: the number of choices for each single choice question, only use it if question_type is "single_choice".
        Defaults to 4.\n
            **output_file_path (str, optional)**: the path of output question set txt file, if not assign, use doc_path+datetime as the file name.
            **kwargs**: the arguments you set in the initial of the class, you can change it here. Include:\n
            question_style, question_type, embeddings, chunk_size, model, verbose, language , search_type, record_exp,
            system_prompt, max_input_tokens, temperature.
        Raises:
            Exception: _description_

        Returns:
            (question_list:list, answer_list:list): the question and answer list that generated by llm model
        """
        ## set class variables ##
        ## set class variables ##
        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.data_source = self._check_doc_path(data_source)
        self.question_num = question_num
        vis_doc_range = set()
        doc_range = (
            (1999 + self.chunk_size) // self.chunk_size
        )  # doc_range is determine by the chunk size, so the select documents won't be too short to having trouble genereating a question

        if check_sum_type(self.question_type, self.question_style):
            return [], []

        self._display_info()
        self._get_db(data_source)
        self._check_db()

        ## set local variables ##
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        start_time = time.time()
        self._add_basic_log(timestamp, "create_questionset", doc_range, choice_num)
        self.prompt_tokens, self.prompt_length = [], []
        self.doc_tokens, self.doc_length = [], []
        self.question, self.answer, self.docs = [], [], []
        table = {}

        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        start_time = time.time()

        self._add_basic_log(timestamp, "create_topic_questionset")

        ## search related documents ##
        retrivers_list = get_retrivers(
            self.db,
            self.embeddings_obj,
            self.threshold,
            self.search_type,
            self.env_file,
        )
        self.docs, doc_length, doc_tokens = search_docs(
            retrivers_list,
            topic,
            self.model,
            self.max_input_tokens - 100,
            self.search_type,
            self.language,
        )

        texts = [doc.page_content for doc in self.docs]
        metadata = [doc.metadata for doc in self.docs]

        doc_range = min(doc_range, len(texts) - 1)

        progress = tqdm(total=question_num, desc=f"Create Q({self.question_type})")
        regenerate_limit = question_num
        ### random select a range of documents from the documents , and use llm model to generate a question and answer pair ###
        for i in range(question_num):
            print(" ")
            progress.update(1)
            print("\n")
            random_index = get_non_repeat_rand_int(
                vis_doc_range, len(texts) - doc_range, doc_range
            )

            doc_text = "\n".join(texts[random_index : random_index + doc_range])
            docs = [
                Document(page_content=texts[k], metadata=metadata[k])
                for k in range(random_index, random_index + doc_range)
            ]
            source_files_name = get_source_files(
                metadata, random_index, doc_range, self.language
            )

            try:
                q_prompt = format_create_question_prompt(
                    doc_text, self.question_type, self.question_style, topic
                )

                response = call_model(self.model_obj, q_prompt)
                if not self._process_response(
                    response, doc_text, choice_num, source_files_name
                ):
                    raise Exception(f"Question Format Error, got {response}")

                self.doc_length.append(get_docs_length(self.language, docs))
                self.doc_tokens.append(self.model_obj.get_num_tokens(doc_text))
                self.prompt_length.append(
                    get_doc_length(self.language, q_prompt) - self.doc_length[-1]
                )
                self.prompt_tokens.append(
                    self.model_obj.get_num_tokens(q_prompt) - self.doc_tokens[-1]
                )
                self.docs.extend(docs)

            except Exception as e:
                if regenerate_limit > 0:
                    regenerate_limit -= 1
                    i -= 1
                    progress.update(-1)
                    logging.warning(f"{e}.\n\n Regenerate\n")
                    continue
                else:
                    logging.error(f"{e}.\n\n Stop\n")
                    raise e

            table = self._update_table(table, docs)

        progress.close()  # end running llm progress bar

        end_time = time.time()

        ### record logs ###
        self._display_info_fnl()
        self._upload_logs(
            end_time - start_time, table, sum(self.doc_length), sum(self.doc_tokens)
        )

        self._add_result_log(timestamp, end_time - start_time)

        self._save_questionset(timestamp, output_file_path)

        return self.question, self.answer

    def evaluation(
        self,
        questionset_file: str,
        data_source: Union[List[str], str],
        eval_model: Union[BaseLanguageModel, str] = "",
        **kwargs,
    ) -> Union[Tuple[float, float, float, list], Tuple[float, list]]:
        """parse the question set txt file generated from "create_questionset" function and then use llm model to generate response,
        evaluate the performance of the given paramters based on similarity between responses and the default answers, use bert_score
        and rouge_l to evaluate the response if you use essay type to generate questionset.  And use correct_count to evaluate
        the response if you use single_choice type to generate questionset.  **Noted that the question_type must match the questionset_file's type**.


            Args:
            **questionset_flie (str)**: the path of question set txt file, accept .txt, .docx and .pdf.\n
            **question_type (str, optional)**: the type of question you want to generate, "essay" or "single_choice". Defaults to "essay".\n
            **eval_model (Union[BaseLanguageModel, str], optional)**: llm model use to score the response. Defaults to "gpt-3.5-turbo".\n
            **kwargs**: the arguments you set in the initial of the class, you can change it here. Include:\n
                embeddings, chunk_size, model, verbose, language , search_type, record_exp,
                system_prompt, max_input_tokens, temperature.\n
        """
        ## set class variables ##
        self._set_model(**kwargs)
        self._change_variables(**kwargs)
        self.data_source = self._check_doc_path(data_source)
        self._decide_eval_model(eval_model)

        ### get question and answer from questionset file ###
        question, answer, self.question_type, self.question_style = (
            get_question_from_file(
                questionset_file, self.question_type, self.question_style
            )
        )

        if check_sum_type(self.question_type, self.question_style):
            if self.question_style.lower() == "essay":
                return 0.0, 0.0, 0.0, []
            else:
                return 0.0, []

        self.system_prompt = (
            check_essay_system_prompt(
                self.question_style, self.language, self.system_prompt
            )
            + self.system_prompt
        )

        ## db ##
        self._display_info()
        self._get_db(data_source)  # create self.db and self.ignore_files
        self._check_db()

        ## set local variables ##
        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        start_time = time.time()
        self.doc_tokens, self.doc_length = [], []
        self.question, self.answer, self.docs = [], [], []
        if self.question_style.lower() == "essay":
            self.score = {"bert": [], "rouge": [], "llm_score": []}
        else:
            self.score = {"correct_count": 0}
        table = {}
        total_docs = []

        self.question_num = len(question)
        progress = tqdm(total=self.question, desc=f"Run Eval({self.question_style})")
        ## add logs ##
        self._add_basic_log(timestamp, "evaluation", -1, -1, questionset_file)

        ### for each question and answer, use llm model to generate response, and evaluate the response by bert_score and rouge_l ###
        retrivers_list = get_retrivers(
            self.db,
            self.embeddings_obj,
            self.threshold,
            self.search_type,
            self.env_file,
        )

        for i in range(self.question_num):
            print(" ")
            progress.update(1)
            print("\n")

            new_table = self._eval_get_res(
                question[i], answer[i], timestamp, retrivers_list
            )
            # ---- #
            if self.record_exp != "":
                for key in new_table:
                    if key not in table:
                        table[key] = []
                    table[key].append(new_table[key])

        progress.close()  # end running llm progress bar
        self.docs = total_docs
        ### record logs ###
        end_time = time.time()
        self.question = question
        self.answer = answer
        self._display_info_fnl()
        self._add_result_log(timestamp, end_time - start_time)

        if self.question_style.lower() == "essay":
            avg_bert = round(sum(self.score["bert"]) / len(self.score["bert"]), 3)
            avg_rouge = round(sum(self.score["rouge"]) / len(self.score["rouge"]), 3)
            avg_llm_score = round(
                sum(self.score["llm_score"]) / len(self.score["llm_score"]), 3
            )

            self._upload_logs(
                end_time - start_time,
                table,
                sum(self.doc_length),
                sum(self.doc_tokens),
                True,
            )

            return avg_bert, avg_rouge, avg_llm_score, self.doc_tokens

        else:
            correct_rate = self.score["correct_count"] / self.question_num
            self._upload_logs(
                end_time - start_time,
                table,
                sum(self.doc_length),
                sum(self.doc_tokens),
                True,
            )

            return correct_rate, self.doc_tokens

    def _decide_eval_model(self, eval_model: Union[BaseLanguageModel, str]):
        if isinstance(eval_model, str):
            if eval_model == "":
                pass
            elif eval_model == self.model:
                self.eval_model = self.model_obj
            else:
                self.eval_model = eval_model
        else:
            self.eval_model = eval_model



from charset_normalizer import detect
from typing import Union
from pathlib import Path
import uuid
import hashlib


def detect_encoding(file_path: Union[str, Path]) -> str:
    with open(file_path, "rb") as file:
        raw_data = file.read(1000)  # Read a portion of the file
        result = detect(raw_data)
        return result["encoding"]


def get_text_md5(text):
    md5_hash = hashlib.md5(text.encode()).hexdigest()

    return md5_hash


def get_mac_address() -> str:
    # Get the MAC address
    mac = uuid.getnode()
    # Convert the MAC address to a readable format without colons
    mac_address = "".join(
        ["{:02x}".format((mac >> elements) & 0xFF) for elements in range(0, 2 * 6, 2)][
            ::-1
        ]
    )
    return get_text_md5(mac_address)



from .upload import aiido_upload
from .prompts.format import (
    handle_language,
    handle_score_table,
    handle_metrics,
    handle_params,
    handle_table,
)
from .search.search_doc import search_docs, retri_docs

__all__ = [
    "aiido_upload",
    "handle_language",
    "handle_score_table",
    "handle_metrics",
    "handle_params",
    "handle_table",
    "search_docs",
    "retri_docs",
]



import streamlit as st
from pathlib import Path


def upload_page():
    """upload documents files to docs folder"""
    st.title("Upload Files")
    st.markdown("##")
    st.markdown("##")

    uploaded_files = st.file_uploader(
        "Upload document files",
        accept_multiple_files=True,
        type=["txt", "pdf", "docx", "md", "csv"],
    )

    path_name = st.text_input("path name")
    sb1, sb2, sb3 = st.columns([1, 1, 1])
    with sb2:
        submit_but = st.button("Submit", type="primary", use_container_width=True)
    if submit_but:
        if path_name in st.session_state.docs_list:
            st.info(
                f"Path name already exist, add files to the path {path_name}.",
                icon="🔥",
            )

        # create path
        if path_name[-1] != "/":
            path_name = path_name + "/"
        save_path = Path(st.session_state.docs_path, path_name)
        # check if path exist
        if not save_path.exists():
            save_path.mkdir(parents=True)
            st.session_state.docs_list.append(path_name[:-1])
        ### write file to doc folder ###
        for uploaded_file in uploaded_files:
            bytes_data = uploaded_file.read()

            with open(save_path.joinpath(uploaded_file.name), "wb") as f:
                f.write(bytes_data)

            st.write("uploaded file:", uploaded_file.name)

        # if len(st.session_state.docs_list) > 0:
        #     st.session_state.chose_doc_path = st.session_state.docs_path + '/'  + st.session_state.docs_list[0]



from typing import Dict, List, Any, Optional, Callable
from pydantic import BaseModel, Field
from langchain.schema.embeddings import Embeddings
from langchain.llms.base import LLM


class custom_embed(BaseModel, Embeddings):
    """HuggingFace sentence_transformers embedding models.

    To use, you should have the ``sentence_transformers`` python package installed.

    Example:
        .. code-block:: python

            from langchain.embeddings import HuggingFaceEmbeddings

            model_name = "sentence-transformers/all-mpnet-base-v2"
            model_kwargs = {'device': 'cpu'}
            encode_kwargs = {'normalize_embeddings': False}
            hf = HuggingFaceEmbeddings(
                model_name=model_name,
                model_kwargs=model_kwargs,
                encode_kwargs=encode_kwargs
            )
    """

    client: Any = Field(default=None)  #: :meta private:
    model_name: str = "custom:custom-embedding-model"
    """Model name to use."""
    # model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    # """Keyword arguments to pass to the model."""
    encode_kwargs: Dict[str, Any] = {}
    """Keyword arguments to pass when calling the `encode` method of the model."""

    def __init__(self, func: Any, encode_kwargs: Dict[str, Any] = {}, **kwargs: Any):
        """Initialize the sentence_transformer."""
        super().__init__(**kwargs)

        self.client = func
        self.encode_kwargs = encode_kwargs
        self.model_name = f"custom:{func.__name__}"

    def embed_documents(self, texts: List[str]) -> List[List[float]]:
        """Compute doc embeddings using a HuggingFace transformer model.

        Args:
            texts: The list of texts to embed.

        Returns:
            List of embeddings, one for each text.
        """
        import numpy

        texts = list(map(lambda x: x.replace("\n", " "), texts))

        embeddings = self.client(texts, **self.encode_kwargs)

        if isinstance(embeddings, numpy.ndarray):
            embeddings = embeddings.tolist()

        return embeddings

    def embed_query(self, text: str) -> List[float]:
        """Compute query embeddings using a HuggingFace transformer model.

        Args:
            text: The text to embed.

        Returns:
            Embeddings for the text.
        """
        return self.embed_documents([text])[0]


class custom_model(LLM):
    max_token: int = 4096
    temperature: float = 0.01
    top_p: float = 0.95
    history: list = []
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)
    func: Any = Field(default=None)

    def __init__(self, func: Callable, temperature: float = 0.001):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__()
        self.func = func
        self.temperature = temperature
        if self.temperature == 0.0:
            self.temperature = 0.001

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return "custom: " + self.func.__name__

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """

        response = self.func(prompt)
        return response



from langchain.tools import BaseTool
from typing import Union, List
import json
import datetime
import time
import logging
import asyncio
import queue
import threading
from langchain_mcp_adapters.client import MultiServerMCPClient
from akasha.utils.atman import basic_llm
from akasha.utils.base import (
    DEFAULT_MODEL,
    DEFAULT_MAX_OUTPUT_TOKENS,
    DEFAULT_MAX_INPUT_TOKENS,
)
from akasha.helper.preprocess_prompts import retri_history_messages
from akasha.helper.base import get_doc_length, extract_json
from akasha.utils.prompts.gen_prompt import format_sys_prompt
from akasha.helper.run_llm import call_model
from .base import (
    get_tool_explaination,
    get_REACT_PROMPT,
    DEFAULT_REMEMBER_PROMPT,
    DEFAULT_OBSERVATION_PROMPT,
    DEFAULT_RETRI_OBSERVATION_PROMPT,
)


class agents(basic_llm):
    """basic class for akasha agent, implement _change_variables, _check_db, add_log and save_logs function."""

    def __init__(
        self,
        tools: Union[BaseTool, List] = [],
        model: str = DEFAULT_MODEL,
        max_input_tokens: int = DEFAULT_MAX_INPUT_TOKENS,
        max_output_tokens: int = DEFAULT_MAX_OUTPUT_TOKENS,
        temperature: float = 0.0,
        prompt_format_type: str = "auto",
        max_round: int = 20,
        max_past_observation: int = 10,
        language: str = "ch",
        record_exp: str = "",
        system_prompt: str = "",
        retri_observation: bool = False,
        keep_logs: bool = True,
        verbose: bool = False,
        stream: bool = False,
        env_file: str = "",
    ):
        """initials of agent class

        Args:
            **model (str, optional)**: llm model to use. Defaults to "gpt-3.5-turbo".\n
            **verbose (bool, optional)**: show log texts or not. Defaults to False.\n
            **language (str, optional)**: the language of documents and prompt, use to make sure docs won't exceed
                max token size of llm input.\n
            **temperature (float, optional)**: temperature of llm model from 0.0 to 1.0 . Defaults to 0.0.\n
            **keep_logs (bool, optional)**: record logs or not. Defaults to False.\n
            ** max_round (int, optional)**: the maximum round of the conversation. Defaults to 20.\n
            ** max_doc_len (int, optional)**: the maximum length of the past thoughts and observations that will send to agent. Defaults to 1500.\n (deprecated in future 1.0.0 version)\n
            ** max_past_observation (int, optional)**: the maximum round of the past thoughts and observations that will send to agent. Defaults to 10.\n
            ** prompt_format_type (str, optional)**: the prompt and system prompt format for the language model, including auto, gpt, llama, chat_gpt, chat_mistral, chat_gemini . Defaults to "auto".
            ** retri_observation (bool, optional)**: if True, agent will ask LLM to retrieve the information from the past thoughts and observations. Defaults to False.\n
            **max_output_tokens (int, optional)**: max output tokens of llm model. Defaults to 1024.\n
            **max_input_tokens (int, optional)**: max input tokens of llm model. Defaults to 3600.\n
        """
        super().__init__(
            model=model,
            max_input_tokens=max_input_tokens,
            max_output_tokens=max_output_tokens,
            temperature=temperature,
            language=language,
            record_exp=record_exp,
            system_prompt=system_prompt,
            keep_logs=keep_logs,
            verbose=verbose,
            env_file=env_file,
        )
        self.stream = stream
        self.prompt_format_type = prompt_format_type
        self.max_round = max_round
        self.max_past_observation = max_past_observation
        self.retri_observation = retri_observation

        ## var ##
        self.messages = []
        self.thoughts = []

        self.tokens = 0
        self.input_len = 0
        self.question = ""
        self.tools = {}
        ## if tools is mcp connection info, connect mcp client to get tools when acall ##

        if isinstance(tools, BaseTool):
            tools = [tools]

        self.tool_explaination = get_tool_explaination(tools)
        for tool in tools:
            if not isinstance(tool, BaseTool):
                logging.warning("tools should be a list of BaseTool")
                continue
            tool_name = tool.name
            self.tools[tool_name] = tool

        self.REACT_PROMPT = self._merge_tool_explaination_and_react()
        self.REMEMBER_PROMPT = DEFAULT_REMEMBER_PROMPT
        self.OBSERVATION_PROMPT = DEFAULT_OBSERVATION_PROMPT
        self.RETRI_OBSERVATION_PROMPT = DEFAULT_RETRI_OBSERVATION_PROMPT

    def _merge_tool_explaination_and_react(self) -> str:
        tool_explain_str = "\n".join(
            [
                f"{tool_name}: {tool_des}"
                for tool_name, tool_des in self.tool_explaination.items()
            ]
        )
        tool_name_str = ", ".join(
            [f'"{tool_name}"' for tool_name in self.tool_explaination.keys()]
        )
        self.tool_name_str = tool_name_str

        return get_REACT_PROMPT(tool_explain_str, tool_name_str)

    def _add_basic_log(self, timestamp: str, fn_type: str) -> bool:
        """add pre-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            fn_type (str): function type of this run
        """
        if super()._add_basic_log(timestamp, fn_type) is False:
            return False

        self.logs[timestamp]["question"] = self.question
        self.logs[timestamp]["max_round"] = self.max_round
        self.logs[timestamp]["max_input_tokens"] = self.max_input_tokens
        self.logs[timestamp]["max_past_observation"] = self.max_past_observation
        return True

    def _add_result_log(self, timestamp: str, time: float) -> bool:
        """add post-process log to self.logs

        Args:
            timestamp (str): timestamp of this run
            time (float): spent time of this run
        """

        if self.keep_logs is False:
            return False

        ### add token information ###
        tool_list = []
        for name, tool in self.tools.items():
            tool_list.append(name)

        self.logs[timestamp]["time"] = time
        self.logs[timestamp]["tools"] = tool_list
        self.logs[timestamp]["messages"] = self.messages
        self.logs[timestamp]["thoughts"] = self.thoughts
        self.logs[timestamp]["response"] = self.response
        self.logs[timestamp]["tokens"] = self.tokens
        self.logs[timestamp]["input_len"] = self.input_len
        return True

    def _display_info(self, batch: int = 1) -> bool:
        """display the information of the parameters if verbose is True"""
        if self.verbose is False:
            return False
        print(f"Model: {self.model}, Temperature: {self.temperature}")
        print("Tool: ", self.tool_name_str)
        print(
            f"Prompt format type: {self.prompt_format_type}, Max input tokens: {self.max_input_tokens}"
        )

        return True

    def __call__(self, question: str, messages: List[dict] = None):
        """Synchronous version of agent."""

        self.question = question

        if not self.stream:
            # Non-streaming: just run and return the result
            async def collect_non_stream():
                return await self._run_agent(
                    question,
                    tool_runner=lambda tool, tool_input: tool._run(**tool_input),
                    messages=messages,
                )

            return asyncio.run(collect_non_stream())

        # Streaming: yield results in real time
        message_queue = queue.Queue()
        stop_event = threading.Event()

        def run_async_stream():
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def process_stream():
                try:
                    async for chunk in self._run_agent_stream(
                        question,
                        tool_runner=lambda tool, tool_input: tool._run(**tool_input),
                        messages=messages,
                    ):
                        message_queue.put(chunk)
                except Exception:
                    import traceback

                    message_queue.put(("ERROR", traceback.format_exc()))
                finally:
                    message_queue.put(None)

            try:
                loop.run_until_complete(process_stream())
            finally:
                pending = asyncio.all_tasks(loop)
                for task in pending:
                    task.cancel()
                if pending:
                    loop.run_until_complete(
                        asyncio.gather(*pending, return_exceptions=True)
                    )
                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.close()

        thread = threading.Thread(target=run_async_stream)
        thread.daemon = True
        thread.start()

        def result_generator():
            try:
                while not stop_event.is_set():
                    try:
                        result = message_queue.get(timeout=0.1)
                        if result is None:
                            break
                        if isinstance(result, tuple) and result[0] == "ERROR":
                            raise RuntimeError(f"Error in async thread: {result[1]}")
                        yield result
                    except queue.Empty:
                        if not thread.is_alive():
                            raise RuntimeError("Background thread died unexpectedly")
            finally:
                stop_event.set()
                if thread.is_alive():
                    thread.join(timeout=5.0)

        return result_generator()

    async def acall(self, question: str, messages: List[dict] = None):
        """Asynchronous version of agent."""

        self.REACT_PROMPT = self._merge_tool_explaination_and_react()

        return await self._run_agent(
            question,
            tool_runner=lambda tool, tool_input: tool.ainvoke(tool_input),
            messages=messages,
        )

    async def _run_agent(
        self, question: str, tool_runner: callable, messages: List[dict] = None
    ):
        """run agent to get response"""
        if self.stream:
            return self._run_agent_stream(question, tool_runner, messages)

        start_time = time.time()
        round_count = self.max_round
        self.response = ""
        if messages is None:
            self.messages = []
        else:
            self.messages = messages
        self.thoughts = []
        observation = ""
        thought = ""
        retri_messages = ""
        self._display_info()
        ### call model to get response ###
        retri_messages, messages_len = retri_history_messages(
            self.messages,
            self.max_past_observation,
            self.max_input_tokens,
            self.model,
            "Action",
            "Observation",
        )
        if retri_messages != "":
            retri_messages = self.OBSERVATION_PROMPT + retri_messages + "\n\n"

        text_input = format_sys_prompt(
            self.REACT_PROMPT,
            "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
            self.prompt_format_type,
            self.model,
        )
        response = call_model(self.model_obj, text_input)

        txt = (
            "Question: "
            + " think step by step"
            + question
            + self.REMEMBER_PROMPT
            + self.REACT_PROMPT
            + retri_messages
        )
        self.input_len = get_doc_length(self.language, txt)
        self.tokens = self.model_obj.get_num_tokens(txt)

        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        if self.keep_logs is True:
            self.timestamp_list.append(timestamp)
            self._add_basic_log(timestamp, "agent_call")

        ### start to run agent ###
        while round_count > 0:
            try:
                cur_action = extract_json(response)
                if (not isinstance(cur_action["action"], str)) or (
                    not isinstance(cur_action["action_input"], dict)
                    and (cur_action["action"] != "Answer")
                ):
                    raise ValueError("Cannot find correct action from response")
            except Exception:
                logging.warning(
                    "Cannot extract JSON format action from response, retry."
                )
                text_input = format_sys_prompt(
                    self.REACT_PROMPT,
                    "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
                    self.prompt_format_type,
                    self.model,
                )
                response = call_model(self.model_obj, text_input)
                round_count -= 1
                txt = (
                    "Question: "
                    + question
                    + retri_messages
                    + self.REACT_PROMPT
                    + self.REMEMBER_PROMPT
                )
                self.input_len += get_doc_length(self.language, txt)
                self.tokens += self.model_obj.get_num_tokens(txt)
                continue

            ### get thought from response ###
            thought = "".join(response.split("Thought:")[1:]).split("Action:")[0]
            if thought.replace(" ", "").replace("\n", "") == "":
                thought = "None."
            self.thoughts.append(thought)

            if cur_action is None:
                raise ValueError("Cannot find correct action from response")
            if cur_action["action"].lower() in [
                "final answer",
                "final_answer",
                "final",
                "answer",
            ]:
                if isinstance(cur_action["action_input"], (dict, list)):
                    response = json.dumps(
                        cur_action["action_input"], ensure_ascii=False
                    )
                else:
                    response = str(cur_action["action_input"])

                self.messages.append(
                    {
                        "role": "Action",
                        "content": json.dumps(cur_action, ensure_ascii=False),
                    }
                )

                break

            elif cur_action["action"] in self.tools:
                try:
                    tool_name = cur_action["action"]
                    tool_input = cur_action["action_input"]

                    tool = self.tools[tool_name]
                    result = tool_runner(tool, tool_input)
                    if asyncio.iscoroutine(result):
                        firsthand_observation = await result  # Await async result
                    else:
                        firsthand_observation = result  # Sync result

                except Exception:
                    logging.warning("Cannot run the tool, retry.")
                    text_input = format_sys_prompt(
                        self.REACT_PROMPT,
                        "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
                        self.prompt_format_type,
                        self.model,
                    )
                    response = call_model(self.model_obj, text_input)
                    round_count -= 1
                    txt = (
                        "Question: "
                        + question
                        + retri_messages
                        + self.REACT_PROMPT
                        + self.REMEMBER_PROMPT
                    )
                    self.input_len += get_doc_length(self.language, txt)
                    self.tokens += self.model_obj.get_num_tokens(txt)
                    continue

                if self.retri_observation:
                    text_input = format_sys_prompt(
                        self.RETRI_OBSERVATION_PROMPT,
                        "Question: "
                        + question
                        + "\n\nThought: "
                        + thought
                        + "\n\nObservation: "
                        + firsthand_observation,
                        self.prompt_format_type,
                        self.model,
                    )
                    observation = call_model(self.model_obj, text_input)
                    txt = (
                        "Question: "
                        + question
                        + "\n\nThought: "
                        + thought
                        + "\n\nObservation: "
                        + firsthand_observation
                        + self.RETRI_OBSERVATION_PROMPT
                    )
                    self.input_len += get_doc_length(self.language, txt)
                    self.tokens += self.model_obj.get_num_tokens(txt)
                else:
                    observation = firsthand_observation

                if self.verbose:
                    print("\nObservation: " + observation)

            else:
                raise ValueError(f"Cannot find tool {cur_action['action']}")

            cur_action["action_input"].pop("run_manager", None)
            self.messages.append(
                {
                    "role": "Action",
                    "content": json.dumps(cur_action, ensure_ascii=False),
                }
            )
            self.messages.append({"role": "Observation", "content": observation})

            retri_messages, messages_len = retri_history_messages(
                self.messages,
                self.max_past_observation,
                self.max_input_tokens,
                self.model,
                "Action",
                "Observation",
            )
            if retri_messages != "":
                retri_messages = self.OBSERVATION_PROMPT + retri_messages + "\n\n"

            text_input = format_sys_prompt(
                self.REACT_PROMPT,
                "Question: " + question + retri_messages,
                self.prompt_format_type,
                self.model,
            )
            response = call_model(self.model_obj, text_input)
            txt = "Question: " + question + retri_messages + self.REACT_PROMPT
            self.input_len += get_doc_length(self.language, txt)
            self.tokens += self.model_obj.get_num_tokens(txt)

            round_count -= 1

        end_time = time.time()
        print(
            "\n-------------------------------------\nSpend Time: ",
            end_time - start_time,
            "s\n",
        )
        self.response = response
        self._add_result_log(timestamp, end_time - start_time)

        return response

    async def _run_agent_stream(
        self, question: str, tool_runner: callable, messages: List[dict] = None
    ):
        """run agent stream to get response"""

        start_time = time.time()
        round_count = self.max_round
        self.response = ""
        if messages is None:
            self.messages = []
        else:
            self.messages = messages
        self.thoughts = []
        observation = ""
        thought = ""
        retri_messages = ""
        self._display_info()
        ### call model to get response ###
        retri_messages, messages_len = retri_history_messages(
            self.messages,
            self.max_past_observation,
            self.max_input_tokens,
            self.model,
            "Action",
            "Observation",
        )
        if retri_messages != "":
            retri_messages = self.OBSERVATION_PROMPT + retri_messages + "\n\n"

        text_input = format_sys_prompt(
            self.REACT_PROMPT,
            "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
            self.prompt_format_type,
            self.model,
        )
        response = call_model(self.model_obj, text_input)

        txt = (
            "Question: "
            + " think step by step"
            + question
            + self.REMEMBER_PROMPT
            + self.REACT_PROMPT
            + retri_messages
        )
        self.input_len = get_doc_length(self.language, txt)
        self.tokens = self.model_obj.get_num_tokens(txt)

        timestamp = datetime.datetime.now().strftime("%Y/%m/%d, %H:%M:%S")
        if self.keep_logs is True:
            self.timestamp_list.append(timestamp)
            self._add_basic_log(timestamp, "agent_call")

        ### start to run agent ###
        while round_count > 0:
            try:
                cur_action = extract_json(response)
                if (not isinstance(cur_action["action"], str)) or (
                    not isinstance(cur_action["action_input"], dict)
                    and (cur_action["action"] != "Answer")
                ):
                    raise ValueError("Cannot find correct action from response")
            except Exception:
                logging.warning(
                    "Cannot extract JSON format action from response, retry."
                )
                text_input = format_sys_prompt(
                    self.REACT_PROMPT,
                    "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
                    self.prompt_format_type,
                    self.model,
                )
                response = call_model(self.model_obj, text_input)
                round_count -= 1
                txt = (
                    "Question: "
                    + question
                    + retri_messages
                    + self.REACT_PROMPT
                    + self.REMEMBER_PROMPT
                )
                self.input_len += get_doc_length(self.language, txt)
                self.tokens += self.model_obj.get_num_tokens(txt)
                continue

            ### get thought from response ###
            thought = "".join(response.split("Thought:")[1:]).split("Action:")[0]
            if thought.replace(" ", "").replace("\n", "") == "":
                thought = "None."
            self.thoughts.append(thought)

            # yield thought #
            intermed_stream_text = "\n[THOUGHT]: " + str(thought) + "\n"
            yield intermed_stream_text

            if cur_action is None:
                raise ValueError("Cannot find correct action from response")
            if cur_action["action"].lower() in [
                "final answer",
                "final_answer",
                "final",
                "answer",
            ]:
                if isinstance(cur_action["action_input"], (dict, list)):
                    response = json.dumps(
                        cur_action["action_input"], ensure_ascii=False
                    )
                else:
                    response = str(cur_action["action_input"])

                self.messages.append(
                    {
                        "role": "Action",
                        "content": json.dumps(cur_action, ensure_ascii=False),
                    }
                )

                yield response

                break

            elif cur_action["action"] in self.tools:
                try:
                    tool_name = cur_action["action"]
                    tool_input = cur_action["action_input"]

                    # yield action #
                    intermed_stream_text = (
                        "\n[ACTION]: "
                        + tool_name
                        + ", "
                        + json.dumps(cur_action, ensure_ascii=False)
                        + "\n"
                    )
                    yield intermed_stream_text

                    tool = self.tools[tool_name]
                    result = tool_runner(tool, tool_input)
                    if asyncio.iscoroutine(result):
                        firsthand_observation = await result  # Await async result
                    else:
                        firsthand_observation = result  # Sync result
                except Exception:
                    logging.warning("Cannot run the tool, retry.")
                    print("Cannot run the tool, retry.")
                    text_input = format_sys_prompt(
                        self.REACT_PROMPT,
                        "Question: " + question + retri_messages + self.REMEMBER_PROMPT,
                        self.prompt_format_type,
                        self.model,
                    )
                    response = call_model(self.model_obj, text_input)
                    round_count -= 1
                    txt = (
                        "Question: "
                        + question
                        + retri_messages
                        + self.REACT_PROMPT
                        + self.REMEMBER_PROMPT
                    )
                    self.input_len += get_doc_length(self.language, txt)
                    self.tokens += self.model_obj.get_num_tokens(txt)
                    continue

                if self.retri_observation:
                    text_input = format_sys_prompt(
                        self.RETRI_OBSERVATION_PROMPT,
                        "Question: "
                        + question
                        + "\n\nThought: "
                        + thought
                        + "\n\nObservation: "
                        + firsthand_observation,
                        self.prompt_format_type,
                        self.model,
                    )
                    observation = call_model(self.model_obj, text_input)
                    txt = (
                        "Question: "
                        + question
                        + "\n\nThought: "
                        + thought
                        + "\n\nObservation: "
                        + firsthand_observation
                        + self.RETRI_OBSERVATION_PROMPT
                    )
                    self.input_len += get_doc_length(self.language, txt)
                    self.tokens += self.model_obj.get_num_tokens(txt)
                else:
                    observation = firsthand_observation

                if self.verbose:
                    print("\n[OBSERVATION]: " + observation)
                # yield observation #
                intermed_stream_text = "\n[OBSERVATION]: " + str(observation) + "\n"
                yield intermed_stream_text

            else:
                raise ValueError(f"Cannot find tool {cur_action['action']}")

            cur_action["action_input"].pop("run_manager", None)
            self.messages.append(
                {
                    "role": "Action",
                    "content": json.dumps(cur_action, ensure_ascii=False),
                }
            )
            self.messages.append({"role": "Observation", "content": observation})

            retri_messages, messages_len = retri_history_messages(
                self.messages,
                self.max_past_observation,
                self.max_input_tokens,
                self.model,
                "Action",
                "Observation",
            )
            if retri_messages != "":
                retri_messages = self.OBSERVATION_PROMPT + retri_messages + "\n\n"

            text_input = format_sys_prompt(
                self.REACT_PROMPT,
                "Question: " + question + retri_messages,
                self.prompt_format_type,
                self.model,
            )
            response = call_model(self.model_obj, text_input)
            txt = "Question: " + question + retri_messages + self.REACT_PROMPT
            self.input_len += get_doc_length(self.language, txt)
            self.tokens += self.model_obj.get_num_tokens(txt)

            round_count -= 1

        end_time = time.time()
        print(
            "\n-------------------------------------\nSpend Time: ",
            end_time - start_time,
            "s\n",
        )
        self.response = response
        self._add_result_log(timestamp, end_time - start_time)

        return

    def mcp_agent(self, connection_info: dict, prompt: str):
        """Call the agent with the given connection info and prompt.

        Args:
            connection_info (dict): _description_
            prompt (str): _description_

        Raises:
            RuntimeError: _description_
            RuntimeError: _description_

        Returns:
            _type_: _description_

        Yields:
            _type_: _description_
        """
        self.question = prompt
        if not self.stream:
            return asyncio.run(self._call_agents_non_streaming(connection_info, prompt))

        # For streaming, use a thread approach with a queue
        import queue
        import threading

        # Create a queue for passing messages from async to sync
        message_queue = queue.Queue()
        stop_event = threading.Event()

        # Function to run in a background thread
        def run_async_stream():
            # Create a new event loop for this thread
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)

            async def process_stream():
                try:
                    client = MultiServerMCPClient(connection_info)
                    tools = await client.get_tools()

                    if isinstance(tools, BaseTool):
                        tools = [tools]

                    self.tool_explaination = get_tool_explaination(tools)
                    for tool in tools:
                        if not isinstance(tool, BaseTool):
                            logging.warning("tools should be a list of BaseTool")
                            continue
                        tool_name = tool.name
                        self.tools[tool_name] = tool

                    # Use the agent asynchronously
                    response = await self.acall(prompt)

                    # Process the streaming response in real-time
                    async for chunk in response:
                        # Put each chunk in the queue as it arrives
                        message_queue.put(chunk)
                except Exception:
                    import traceback

                    message_queue.put(("ERROR", traceback.format_exc()))
                finally:
                    # Signal that we're done
                    message_queue.put(None)

            # Run the async function and ensure proper cleanup
            try:
                loop.run_until_complete(process_stream())
            finally:
                # Clean up the event loop
                pending = asyncio.all_tasks(loop)
                for task in pending:
                    task.cancel()

                # Give cancelled tasks a chance to clean up
                if pending:
                    loop.run_until_complete(
                        asyncio.gather(*pending, return_exceptions=True)
                    )

                loop.run_until_complete(loop.shutdown_asyncgens())
                loop.close()

        # Start the background thread
        thread = threading.Thread(target=run_async_stream)
        thread.daemon = True  # The thread will exit when the main thread exits
        thread.start()

        # Return a generator that yields results as they arrive from the queue
        def result_generator():
            try:
                while not stop_event.is_set():
                    try:
                        # Timeout allows checking the stop_event occasionally
                        result = message_queue.get(timeout=0.1)

                        # None signals end of stream
                        if result is None:
                            break

                        # Check for error
                        if isinstance(result, tuple) and result[0] == "ERROR":
                            raise RuntimeError(f"Error in async thread: {result[1]}")

                        yield result

                    except queue.Empty:
                        # Just a timeout, check if the thread is still alive
                        if not thread.is_alive():
                            # Thread died unexpectedly
                            raise RuntimeError("Background thread died unexpectedly")
                        # Otherwise continue waiting
            finally:
                # Clean up
                stop_event.set()
                # Wait for thread to finish if it's still running
                if thread.is_alive():
                    thread.join(timeout=5.0)

        return result_generator()

    ## use MultiServerMCPClient to connect to multiple MCP servers and get the tools
    async def _call_agents_non_streaming(self, connection_info: dict, prompt: str):
        """Handle the non-streaming case where we want to return a complete string"""

        client = MultiServerMCPClient(connection_info)
        tools = await client.get_tools()

        if isinstance(tools, BaseTool):
            tools = [tools]

        self.tool_explaination = get_tool_explaination(tools)
        for tool in tools:
            if not isinstance(tool, BaseTool):
                logging.warning("tools should be a list of BaseTool")
                continue
            tool_name = tool.name
            self.tools[tool_name] = tool

        # Use the agent asynchronously
        response = await self.acall(prompt)

        # For non-streaming, collect the complete response
        if hasattr(response, "__aiter__"):
            # If it's an async generator, collect all chunks and join them
            result = ""
            async for chunk in response:
                result += chunk
            return result
        # If it's already a string or other non-generator response
        return response



from .create_db import create_single_file_db, create_directory_db, create_webpage_db  # noqa: F401
from .db_structure import dbs, get_storage_directory  # noqa: F401
from .load_db import process_db, load_db_by_chroma_name  # noqa: F401
from .load_docs import load_docs_from_info  # noqa: F401
from .delete_db import delete_documents_by_file, delete_documents_by_directory  # noqa: F401
from .extract_db import (
    extract_db_by_file,  # noqa: F401
    extract_db_by_ids,  # noqa: F401
    extract_db_by_keyword,  # noqa: F401
    pop_db_by_ids,  # noqa: F401
)
from .upadte_db import update_db  # noqa: F401



from typing import List, Dict, Callable, Union
import logging
import inspect
from langchain.tools import BaseTool


def get_tool_explaination(tools: List[BaseTool]) -> Dict[str, str]:
    """get the explaination and parameter name, type of tools"""
    ret = {}
    for tool in tools:
        name = tool.name
        dscp = tool.description
        para_des = ""
        try:
            para_des = tool.para_des
        except Exception:
            try:
                para_des = "args: {"
                dic = tool.args
                for k, v in dic.items():
                    para_des += f"{k}:{v['type']}, "
                para_des = para_des[:-2]  # remove the last comma and space
                para_des += "}"

            except Exception:
                para_des = ""
        ret[name] = dscp + ", " + para_des

    return ret


def create_tool(
    tool_name: str, tool_description: str, func: Callable
) -> Union[BaseTool, None]:
    """input function to create a tool

    Args:
        tool_name (str): name of the tool
        tool_description (str): description of the function and the parameters of the tool
        func (Callable): callable function of the tool

    Returns:
        Union[BaseTool, None]: return the tool if success, return None if failed
    """
    try:

        class custom_tool(BaseTool):
            name: str
            description: str
            para_des: str

            async def ainvoke(self, tool_input: dict, **kwargs):
                """Asynchronous invocation of the tool."""
                if inspect.iscoroutinefunction(func):
                    # If func is asynchronous, await it
                    return await func(**tool_input, **kwargs)
                else:
                    # If func is synchronous, run it directly
                    return func(**tool_input, **kwargs)

            def _run(self, *args, **kwargs):
                return func(*args, **kwargs)

            def run(self, *args, **kwargs):
                return self._run(*args, **kwargs)

        sig = inspect.signature(func)
        params = sig.parameters

        prm = "args: {"
        for n, p in params.items():
            if p.annotation is not inspect.Parameter.empty:
                prm += f" {n}: {p.annotation},"
        prm += "}"

    except Exception as e:
        logging.error(f"Cannot create tool correctly, {e}\n\n")
        raise e

    return custom_tool(name=tool_name, description=tool_description, para_des=prm)


def get_REACT_PROMPT(tool_explain_str: str, tool_name_str: str) -> str:
    """get the REACT prompt for the tool"""

    ret = f"""Respond to the human as helpfully and accurately as possible. You have access to the following tools:\n\n{tool_explain_str}\n
Use a json blob to specify a tool by providing an action key (tool name) and an action_input key (tool input).\n\n
Valid "action" values: "Answer" or {tool_name_str}\n\n
Provide only ONE action per $JSON_BLOB, as shown:\n{{\n```\n\n  "action": $TOOL_NAME,\n  "action_input": $INPUT\n}}\n```$INPUT is a dictionary that contains tool parameters and their values\n\n
the meaning of each format:\n
Question: input question to answer\nThought: consider previous and subsequent steps\nAction:\n```\n$JSON_BLOB\n```\nObservation: action result\n
... (repeat Thought/Action N times)\nThought: I know what to respond\nAction:\n```\n{{\n  "action": "Answer",\n  "action_input": "Final response to human"\n}}\n```\n\n
Begin! Reminder to ALWAYS respond with a valid json blob of a single action. Use tools if necessary. Respond directly if appropriate. Format is Thought: then Action:```$JSON_BLOB```.\n
"""

    return ret


DEFAULT_REMEMBER_PROMPT = (
    "**Remember, Format is Thought: then Action:```$JSON_BLOB```\n\n"
)


DEFAULT_OBSERVATION_PROMPT = "\n\nBelow are your previous work, check them carefully and provide the next action and thought,**do not ask same question repeatedly: "
DEFAULT_RETRI_OBSERVATION_PROMPT = "User will give you Question, Thought and Observation, return the information from Observation that you think is most relevant to the Question or Thought, if you can't find the information, return None."



from langchain.schema import Document
from typing import Union, List
from pathlib import Path
from akasha.utils.db.db_structure import is_url
from akasha.utils.db.file_loader import load_file, load_directory, load_url


def _get_file_dir_docs(s: Union[str, Path]) -> List[Document]:
    try:
        if isinstance(s, str):
            if len(s) < 255:
                s = Path(s)
            else:
                return [Document(page_content=s)]
    except Exception as e:
        print(f"Error loading {s}: {e}")
        return [Document(page_content=s)]

    try:
        if s.is_file():
            return load_file(s, s.__str__().split(".")[-1])
        elif s.is_dir():
            return load_directory(s)
        else:
            return [Document(page_content=s.__str__())]
    except Exception as e:
        print(f"Error loading {s}: {e}")
        return [Document(page_content=s.__str__())]


def load_docs_from_info(info: Union[str, list, Path, Document]) -> List[Document]:
    """_summary_

    Args:
        info (Union[str, list, Path, Document]): _description_

    Returns:
        List[Document]: _description_
    """
    docs = []
    if not isinstance(info, list):
        info = [info]

    for si in info:
        if isinstance(si, str):
            if si == "":
                continue
            if is_url(si):
                docs.append(load_url(si))
            else:
                docs.extend(_get_file_dir_docs(si))
        elif isinstance(si, Path):
            docs.extend(_get_file_dir_docs(si))
        elif isinstance(si, Document):
            docs.append(si)

    return docs



from typing import List, Any, Optional, Generator, Union
from langchain.llms.base import LLM
from transformers import AutoTokenizer, AutoModelForCausalLM, TextIteratorStreamer

import torch
import requests
from threading import Thread

from PIL import Image
from pydantic import Field


class hf_model(LLM):
    max_token: int = 4096
    tokenizer: Any = Field(default=None)
    model: Any = Field(default=None)
    streamer: Any = Field(default=None)
    device: Any = Field(default=None)
    model_id: str
    processor: Any = None
    temperature: float = 0.01
    hf_token: Union[str, None] = None
    max_output_tokens: int = 1024
    model_name: str = ""

    def __init__(self, model_name: str, env_dict: dict, temperature: float, **kwargs):
        """define custom model, input func and temperature

        Args:
            **func (Callable)**: the function return response from llm\n
        """
        super().__init__(model_id=model_name)
        if "HF_TOKEN" in env_dict:
            self.hf_token = env_dict["HF_TOKEN"]
        elif "HUGGINGFACEHUB_API_TOKEN" in env_dict:
            self.hf_token = env_dict["HUGGINGFACEHUB_API_TOKEN"]
        else:
            self.hf_token = None

        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        if temperature == 0.0:
            temperature = 0.01
        if "max_output_tokens" in kwargs:
            self.max_output_tokens = kwargs["max_output_tokens"]

        self.temperature = temperature
        self.model_name = model_name
        if "vision" in model_name.lower():
            self.init_vision_model()
        else:
            self.tokenizer = AutoTokenizer.from_pretrained(
                model_name, token=self.hf_token
            )
            self.streamer = TextIteratorStreamer(self.tokenizer, skip_prompt=True)
            self.model = AutoModelForCausalLM.from_pretrained(
                model_name,
                token=self.hf_token,
                temperature=temperature,
                repetition_penalty=1.2,
                top_p=0.95,
                torch_dtype=torch.float16,
                device_map="auto",
            )

    @property
    def _llm_type(self) -> str:
        """return llm type

        Returns:
            str: llm type
        """
        return f"hf:{self.model_name} huggingface"

    def init_vision_model(self, **kwargs):
        """init vision model

        Args:
            model_name (str): model name
        """
        try:
            from transformers import MllamaForConditionalGeneration, AutoProcessor

            self.model = MllamaForConditionalGeneration.from_pretrained(
                self.model_id,
                torch_dtype=torch.bfloat16,
                device_map="auto",
            )
            self.processor = AutoProcessor.from_pretrained(self.model_id)
        except Exception:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)
            self.streamer = TextIteratorStreamer(self.tokenizer, skip_prompt=True)
            self.model = AutoModelForCausalLM.from_pretrained(
                self.model_id,
                token=self.hf_token,
                temperature=self.temperature,
                repetition_penalty=1.2,
                top_p=0.95,
                torch_dtype=torch.float16,
                device_map="auto",
            )

    def stream(
        self, prompt: str, stop: Optional[List[str]] = None
    ) -> Generator[str, None, None]:
        stop_list = get_stop_list(stop)
        # inputs = self.tokenizer([prompt], return_tensors="pt").to(self.device)
        inputs = self.tokenizer([prompt], return_tensors="pt")
        inputs = {k: v.to(self.device) for k, v in inputs.items()}
        generation_kwargs = dict(
            inputs,
            streamer=self.streamer,
            max_new_tokens=self.max_output_tokens,
            do_sample=True,
            min_new_tokens=10,
            stop_strings=stop_list,
            tokenizer=self.tokenizer,
        )
        # self.model.generate(**inputs, streamer= self.streamer, max_new_tokens=1024, do_sample=True)

        thread = Thread(target=self.model.generate, kwargs=generation_kwargs)
        thread.start()
        # for text in self.streamer:
        #     yield text
        yield from self.streamer

    def _call(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        """run llm and get the response

        Args:
            **prompt (str)**: user prompt
            **stop (Optional[List[str]], optional)**: not use. Defaults to None.\n

        Returns:
            str: llm response
        """
        # inputs = self.tokenizer([prompt], return_tensors="pt").to(self.device)
        inputs = self.tokenizer([prompt], return_tensors="pt")
        inputs = {k: v.to(self.device) for k, v in inputs.items()}
        stop_list = get_stop_list(stop)
        gerneration_kwargs = dict(
            inputs,
            streamer=self.streamer,
            max_new_tokens=self.max_output_tokens,
            do_sample=True,
            stop_strings=stop_list,
            tokenizer=self.tokenizer,
        )

        thread = Thread(target=self.model.generate, kwargs=gerneration_kwargs)
        thread.start()
        generated_text = ""
        for new_text in self.streamer:
            print(new_text, end="", flush=True)
            generated_text += new_text
        return generated_text

    def _generate(self, prompt: str, stop: Optional[List[str]] = None) -> str:
        return self._call(prompt, stop)

    def batch(self, prompt: List[str], stop: Optional[List[str]] = None) -> List[str]:
        self.tokenizer.pad_token = self.tokenizer.eos_token

        # Tokenize and move to correct device
        inputs = self.tokenizer(
            prompt, return_tensors="pt", padding=True, truncation=True, max_length=1024
        )
        inputs = {k: v.to(self.device) for k, v in inputs.items()}

        # Note: `stop_strings` is not supported by HF `generate()`, may require custom logic
        generated_ids = self.model.generate(
            **inputs,
            max_new_tokens=self.max_output_tokens,
            do_sample=True,
            # stop_strings=stop_list  ← only keep if your model wrapper supports it
        )

        generated_texts = self.tokenizer.batch_decode(
            generated_ids, skip_special_tokens=True
        )

        return generated_texts

    def call_image(self, prompt: list) -> str:
        def is_url(path):
            from urllib.parse import urlparse

            parsed_url = urlparse(path)
            return parsed_url.scheme in ("http", "https", "ftp")

        ### check if the image is from url or local path
        image_path = prompt[0]["content"][0]["type"]
        prompt[0]["content"][0]["type"] = "image"

        if is_url(image_path):
            image = Image.open(requests.get(image_path, stream=True).raw)
        else:
            image = Image.open(image_path)

        ## apply chat template ##
        input_text = self.processor.apply_chat_template(
            prompt, add_generation_prompt=True
        )
        inputs = self.processor(
            image, input_text, add_special_tokens=False, return_tensors="pt"
        ).to(self.device)

        output = self.model.generate(
            **inputs,
            max_new_tokens=512,
            do_sample=True,
            temperature=self.temperature,
            top_p=0.95,
            length_penalty=1.0,
            repetition_penalty=1.0,
        )
        return self.processor.decode(
            output[0][len(prompt[0]["content"][1]["text"]) + 8 :]
        )


def get_stop_list(stop: Optional[List[str]]) -> List[str]:
    """get stop list

    Args:
        stop (Optional[List[str]]): stop list

    Returns:
        List[str]: stop list
    """
    ret = ["<|eot_id|>", "<|end_header_id|>", "</s>"]
    if stop is not None:
        ret = stop
    return ret



from pathlib import Path
import os

from akasha.helper.base import separate_name


class myTokenizer(object):
    """this class is for computing the number of tokens in a given text using different tokenizers.

    Args:
        object (_type_): _description_
    """

    def __init__(self, model_id: str, tokenizer: object, path: str = "./tokenizers"):
        """
        Initialize a Tokenizer object.

        Args:
            model_id (str): The name of the model for the tokenizer.
            tokenizer (object): The tokenizer object from HuggingFace.
            path (str, optional): The path to save the tokenizer. Defaults to './tokenizers'.
        """
        self.model_id = model_id
        self.tokenizer = tokenizer
        self.path = path

    @classmethod
    def load(cls, model_id: str, path: str = "./tokenizers"):
        """
        Load a tokenizer from local path or huggingface model hub.

        Args:
            model_id (str): The name of the model. only supported on Non-OpenAI models & gpt-2 model.
                            ex. 'google/gemma-2-2b-it', 'openai:gpt2'
                            Reminder: model_id which includes '/' will be replaced by '--' to find the local path
            path (str, optional): The path to the tokenizer. Defaults to './tokenizers'.
        """
        from transformers import AutoTokenizer

        model_path = Path.joinpath(Path(path), Path(model_id.replace("/", "--")))
        if model_path.exists():
            print("Loading tokenizer from local path...")
            tokenizer = AutoTokenizer.from_pretrained(model_path)
        else:
            print("Loading tokenizer from huggingface model hub...")
            hf_token = os.environ.get("HF_TOKEN")
            if hf_token is None:
                hf_token = os.environ.get("HUGGINGFACEHUB_API_TOKEN")
            tokenizer = AutoTokenizer.from_pretrained(model_id, token=hf_token)
        return cls(model_id, tokenizer, path)

    def save(self, name: str, path: str = None):
        """
        Save the tokenizer to local path.

        Args:
            name (str): The name of the tokenizer to be saved.
            path (str, optional): The path to the tokenizer. Defaults to the path given in `__init__`.
        """
        if self.tokenizer is None:
            raise ValueError("Tokenizer is not initialized")
        if path is None:
            path = self.path
        root_path = Path(path)
        if not root_path.exists():
            Path.mkdir(root_path)
        model_path = Path.joinpath(root_path, Path(name))
        if not model_path.exists():
            Path.mkdir(model_path)
        self.tokenizer.save_pretrained(model_path)

    def compute_tokens_huggingface(self, text: str) -> int:
        """
        Compute the number of tokens in a given text using huggingface tokenizer.

        Args:
            text (str): The text to be tokenized.

        Returns:
            int: The number of tokens in the text.
        """
        tokens = self.tokenizer(text)
        num_tokens = len(tokens["input_ids"])
        return num_tokens

    @staticmethod
    def compute_tokens_anthropic(text: str, model_name: str) -> int:
        ### take too long time to load the model, skip for now ###
        import anthropic

        token_count = anthropic.Anthropic().beta.messages.count_tokens(
            model=model_name, messages=[{"role": "user", "content": text}]
        )

        return token_count.input_tokens

    @staticmethod
    def compute_tokens_openai(text: str, model_name: str) -> int:
        """
        Compute the number of tokens in a given text using OpenAI tiktoken.

        Args:
            text (str): The text to be tokenized.
            model_name (str): The name of the OpenAI model. ex. 'openai:gpt-3.5-turbo'
                            Reminder: '/' should not be included in model_name

        Returns:
            int: The number of tokens in the text.

        Raises:
            ValueError: If the model_name is not a valid OpenAI model.
        """
        import tiktoken

        if "/" in model_name:
            raise ValueError("Non-OpenAI models are not supported")
        if model_name.lower().startswith("openai:"):
            model_name = model_name.lower().lstrip("openai:")
        encoding = tiktoken.encoding_for_model(model_name)
        tokens = encoding.encode(text)
        num_tokens = len(tokens)
        return num_tokens

    @staticmethod
    def compute_tokens_gemini(text: str, model_name: str) -> int:
        """
        Compute the number of tokens in a given text using Google Vertex AI.

        Args:
            text (str): The text to be tokenized.
            model_name (str): The name of the Gemini model. ex. 'gemini:gemini-1.5-flash'


        Returns:
            int: The number of tokens in the text.

        Raises:
            ValueError: If the model_id is not a valid Gemini model.
        """
        try:
            from akasha.utils.models.gemi import calculate_token

            if "/" in model_name:
                model_name = model_name.split("/")[-1]
            if model_name.lower().startswith("gemini:"):
                model_name = model_name.lower().lstrip("gemini:")
            # tokenizer = tokenization.get_tokenizer_for_model(model_name)
            # num_tokens = tokenizer.count_tokens(text).total_tokens
            num_tokens = calculate_token(text, model_name=model_name)
        except Exception:
            import tiktoken

            encoding = tiktoken.get_encoding("cl100k_base")
            num_tokens = len(encoding.encode(text))
            return num_tokens

        return num_tokens

    @classmethod
    def compute_tokens(
        cls,
        text: str,
        model_id: str,
        model_path: str = "./tokenizers",
        save_tokenizer: bool = True,
    ) -> int:
        """
        Compute the number of tokens in a given text using either huggingface or OpenAI tiktoken.

        If the model_id is an OpenAI model, the tiktoken library is used to tokenize the text.
        If the model_id is a Non-OpenAI model, the huggingface tokenizer is used to tokenize the text.

        Args:
            text (str): The text to be tokenized.
            model_id (str): The name of the model. ex. 'gpt-2', 'openai:gpt-3.5-turbo'
            model_path (str, optional): The path to the tokenizer. Defaults to './tokenizers'.
            save_tokenizer (bool, optional): Whether to save the tokenizer locally. Defaults to True.

        Returns:
            int: The number of tokens in the text.

        """
        model_type, model_name = separate_name(model_id)
        if model_type in ["openai", "gpt-3.5", "gpt"]:
            return cls.compute_tokens_openai(text, model_name)
        elif model_type in ["gemini", "google"]:
            return cls.compute_tokens_gemini(text, model_name)
        elif model_type in [
            "huggingface",
            "huggingfacehub",
            "transformers",
            "transformer",
            "huggingface-hub",
            "hf",
        ]:
            tkn = cls.load(model_name, model_path)
            # tokenizer is storable if using Non-OpenAI model
            if save_tokenizer:
                tkn.save(name=model_name.replace("/", "--"), path=model_path)
            return tkn.compute_tokens_huggingface(text)

        else:
            import tiktoken

            encoding = tiktoken.get_encoding("cl100k_base")
            num_tokens = len(encoding.encode(text))
            return num_tokens



import streamlit as st
import akasha
import sys

sys.path.append("../")


def response_page():
    """implement rag ui"""

    if "prompt_list" not in st.session_state:
        st.session_state.prompt_list = []

    st.title("RAG")
    response_board, para_set = st.columns([3, 1])

    with para_set:
        st.session_state.sys_prompt = st.text_area(
            "System Prompt",
            st.session_state.sys_prompt,
            help="The special instruction you want to give to the model.",
        )

        sb1, sb2 = st.columns([1, 1])
        with sb1:
            if st.button(
                "Clear",
                type="primary",
                use_container_width=True,
                help="Clear the prompt and response history.",
            ):
                st.session_state.prompt_list = []
                st.session_state.response_list = []
                st.rerun()
        with sb2:
            st.session_state.stream = st.toggle(
                "Stream",
                value=st.session_state.stream,
                help="the output to be stream or not.",
            )

    for i in range(len(st.session_state.response_list)):
        with response_board.chat_message("user"):
            st.markdown(st.session_state.prompt_list[i])
        with response_board.chat_message("assistant"):
            st.markdown(st.session_state.response_list[i])

    prompt_borad, para_set2 = st.columns([3, 1])
    with prompt_borad:
        prompt = st.chat_input(
            "Ask your question here",
        )

    if prompt:
        ## check if the object is created correctly ##
        with response_board.chat_message("user"):
            st.markdown(prompt)
        if not isinstance(st.session_state.akasha_obj, akasha.RAG):
            st.session_state.akasha_obj = akasha.RAG(
                embeddings=st.session_state.embed,
                chunk_size=st.session_state.chunksize,
                model=st.session_state.model,
                search_type=st.session_state.search_type,
                language="ch",
                verbose=True,
                record_exp="",
                max_input_tokens=st.session_state.max_input_tokens,
                temperature=st.session_state.temperature,
                env_file=st.session_state.env_path,
            )

        ans = st.session_state.akasha_obj(
            st.session_state.chose_doc_path,
            prompt,
            embeddings=st.session_state.embed,
            chunk_size=st.session_state.chunksize,
            model=st.session_state.model,
            search_type=st.session_state.search_type,
            system_prompt=st.session_state.sys_prompt,
            max_input_tokens=st.session_state.max_input_tokens,
            temperature=st.session_state.temperature,
            keep_logs=True,
            env_file=st.session_state.env_path,
            stream=st.session_state.stream,
        )

        with response_board.chat_message("assistant"):
            if st.session_state.stream:
                ans = st.write_stream(ans)
            else:
                st.markdown(ans)

        st.session_state.prompt_list.append(prompt)
        st.session_state.response_list.append(ans)
        timesp = st.session_state.akasha_obj.timestamp_list[-1]
        if timesp in st.session_state.akasha_obj.logs:
            st.session_state.logs[timesp] = st.session_state.akasha_obj.logs[timesp]



from typing import Callable, Union, List
from langchain.tools import BaseTool
from akasha.helper.base import extract_json
from .base import create_tool
import json


def websearch_tool(
    search_engine: str = "wiki", language: str = "ch", env_file: str = ""
) -> BaseTool:
    """return the tool to use search engine to search information for user

    Args:
        search_engine (str, optional): the search_engine options, includes "wiki", "serper" "tavily" and "brave". Defaults to "wiki".
        language (str, optional): the language options, includes "ch" and "en". Defaults to "ch".
        env_file (str, optional): the .env file save the api keys. Defaults to "".

    Returns:
        _type_: _description_
    """
    from akasha.helper.web_engine import load_docs_from_webengine

    def _websearch_tool(prompt: str, search_num: int = 5):
        docs = load_docs_from_webengine(
            prompt, search_engine, search_num, language, env_file
        )

        ret = "\n\n".join([do.page_content for do in docs])
        return ret

    ret_tool = create_tool(
        tool_name="web_search_tool",
        tool_description="This is the tool to use prompt to search information from web engine, the params \
        includes prompt, search_num(default 5)",
        func=_websearch_tool,
    )

    return ret_tool


def calculate_tool():
    def _cal_tool(expr: str):
        return eval(expr)

    ret_tool = create_tool(
        tool_name="calculate_tool",
        tool_description="This is the tool to calculate the math expression, the only one parameter is expr",
        func=_cal_tool,
    )

    return ret_tool


def saveJSON_tool() -> BaseTool:
    """return the json save tool that can save the content into json file.

    Returns:
        _type_: _description_
    """
    return create_tool(
        tool_name="json_tool",
        tool_description="This is the tool to save the content into json file, the input contains file_path and content.",
        func=_jsonSaveTool,
    )


def _jsonSaveTool(file_path: str = "default.json", content: str = None):
    """save content into json file"""
    if content:
        try:
            # change content from string to json
            try:
                content = extract_json(content)
            except Exception:
                print(content)
            import json

            with open(file_path, "w", encoding="utf-8") as f:
                json.dump(content, f, indent=4, ensure_ascii=False)
            return f"Success create {file_path}"
        except Exception as e:
            print("content: ", content)
            return f"{e}, Cannot save file_path {file_path}, save file as default.json"
    else:
        return "Error: content is empty"


def rag_tool(
    embeddings: Union[str, Callable],
    chunk_size: int = 1000,
    max_input_tokens: int = 3000,
    env_file: str = "",
):
    from akasha.utils.db import process_db
    from akasha.helper.handle_objects import handle_embeddings
    from akasha.utils.search.retrievers.base import get_retrivers
    from akasha.utils.search.search_doc import search_docs

    embeddings_obj = handle_embeddings(embeddings, env_file=env_file)

    def _rag_tool(
        data_source: Union[List[str], str], prompt: str, search_type: str = "knn"
    ):
        # Ensure data_source is a list
        if isinstance(data_source, str):
            try:
                # Attempt to load as JSON
                data_source = json.loads(data_source)
                if not isinstance(data_source, list):
                    # If it's not a list, treat it as a single path string
                    data_source = [data_source]
            except json.JSONDecodeError:
                # If JSON decoding fails, treat it as a single path string
                data_source = [data_source]
        elif not isinstance(data_source, list):
            raise ValueError(
                f"data_source should be a list of strings or a single string, instead, we got {data_source}."
            )

        db = process_db(
            data_source=data_source,
            embeddings=embeddings_obj,
            chunk_size=chunk_size,
            env_file=env_file,
        )

        retrivers_list = get_retrivers(
            db,
            embeddings_obj,
            0.0,
            search_type,
            env_file,
        )

        docs, doc_length, doc_tokens = search_docs(
            retrivers_list,
            prompt,
            max_input_tokens=max_input_tokens,
            search_type=search_type,
        )

        return "\n\n".join([do.page_content for do in docs])

    ret_tool = create_tool(
        tool_name="rag_tool",
        tool_description="This is the tool to use prompt to search information from db, the params \
        includes data_source(str or list, the path of data source), prompt, search_type(default 'knn')",
        func=_rag_tool,
    )

    return ret_tool



import requests
from bs4 import BeautifulSoup
from typing import Tuple
from datetime import datetime
import time


def get_text_from_url(url: str) -> Tuple[str, str]:
    try:
        # Send a GET request to the URL
        response = requests.get(url)
        response.raise_for_status()  # Raise an exception for HTTP errors

        # Parse the HTML content
        soup = BeautifulSoup(response.content, "html.parser")

        # Get the title of the web page
        title = soup.title.string if soup.title else "No title found"

        # Remove unwanted elements
        for element in soup(
            [
                "meta",
                "footer",
                "nav",
                "a",
            ]
        ):
            element.decompose()

        # Extract and return the text content
        text_content = soup.get_text(separator=" ", strip=True)
        return title, text_content
    except requests.RequestException as e:
        print(f"Error fetching the URL: {e}")
        return "", ""


def get_webpage_last_modified(url):
    try:
        # Send a HEAD request to get headers
        response = requests.head(url)
        response.raise_for_status()  # Raise an exception for HTTP errors

        # Get the Last-Modified header
        last_modified = response.headers.get("Last-Modified")
        if last_modified:
            # Convert to datetime object
            last_modified_date = datetime.strptime(
                last_modified, "%a, %d %b %Y %H:%M:%S %Z"
            )
            # Convert to timestamp (float)
            last_modified_timestamp = time.mktime(last_modified_date.timetuple())
            return last_modified_date, last_modified_timestamp
        else:
            return None
    except requests.RequestException as e:
        print(f"Error fetching the URL: {e}")
        return None
